/// @file
////////////////////////////////////////////////////////////////////////////////
///
/// Copyright (C) 2009      University of Toronto
///
////////////////////////////////////////////////////////////////////////////////
///
///  module     : condition number optimization
///
///  author     : lessig@dgp.toronto.edu
///
///  project    : Spherical Harmonics rotation
///
///  description: computation of "energy" for optimization
///
///  license    : See README for details.
///
////////////////////////////////////////////////////////////////////////////////

// includes, file
#include <energy.h>

// includes, system
#include <cstring>
#include <cstdio>

// Lapack SVD routine 
extern "C"
int 
dgesvd_(char *jobu, char *jobvt, int *m, int *n, 
	      double *a, int *lda, double *s, double *u, int* ldu,
	      double *vt, int *ldvt, double *work, int *lwork, 
      	int *info);

////////////////////////////////////////////////////////////////////////////////
//! Constructor
////////////////////////////////////////////////////////////////////////////////
Energy::Energy() :
  sh_band( 0),
  sh_band_dim( 0),
  num_samples( 0),
  fd_eps_theta( 0.0),
  fd_eps_phi( 0.0),
  svals( 0),
  q_data( NULL),
  ws( NULL),
  ws_size( 0),
  sh_eval( ShEvaluator<double>::instance()),
  sh_vals( 0)
{ }

////////////////////////////////////////////////////////////////////////////////
//! Destructor
////////////////////////////////////////////////////////////////////////////////
Energy::~Energy() {

  // clean up
  if( NULL != q_data) {
    delete[] q_data;
    delete[] q_trans;
  }

  if( NULL != ws) {
    delete[] ws;
  }
}

////////////////////////////////////////////////////////////////////////////////
//! Initialization
////////////////////////////////////////////////////////////////////////////////
void
Energy::init( const unsigned int n_samples, const unsigned int band, 
              const double eps_theta, const double eps_phi) 
{
  sh_band = band;
  sh_band_dim = 2*sh_band+1;

  num_samples = n_samples;

  fd_eps_theta = eps_theta;
  fd_eps_phi = eps_phi;

  svals.resize( sh_band_dim);
  sh_vals.resize( sh_band_dim);

  q_data = new double[sh_band_dim * num_samples];
  q_trans = new double[sh_band_dim * num_samples];

  sh_eval->init( sh_band);

  ws_size = 12*num_samples;
  ws = new double[ws_size];
};

////////////////////////////////////////////////////////////////////////////////
//! Compute energy and gradient at \a lambdas
////////////////////////////////////////////////////////////////////////////////
void
Energy::compute( const LambdaVec& lambdas, double& energy, Vec& gradient) {

  // compute energy
  energy = computeCondNum( lambdas);

  // helper variables for computing the gradient
  LambdaVec lambdas_delta_theta( lambdas);
  LambdaVec lambdas_delta_phi( lambdas);
  double cn_delta = 0.0;

  // compute gradient
  for( unsigned int i = 0; i < num_samples; ++i) {

    lambdas_delta_theta = lambdas;
    lambdas_delta_theta[i].first += fd_eps_theta;

    cn_delta = computeCondNum( lambdas_delta_theta);
    gradient[2*i] = (cn_delta - energy) / fd_eps_theta;

    lambdas_delta_phi = lambdas;
    lambdas_delta_phi[i].second += fd_eps_phi;

    cn_delta = computeCondNum( lambdas_delta_phi);
    gradient[2*i+1] = (cn_delta - energy) / fd_eps_phi;
  }
}

////////////////////////////////////////////////////////////////////////////////
//! Transpose kernel matrix to match Fortran format used in Lapack
////////////////////////////////////////////////////////////////////////////////
void 
Energy::transposeQ() {

  for( unsigned int i = 0; i < num_samples; ++i) {
    for( unsigned int j = 0; j < sh_band_dim; ++j) {
      q_trans[j*num_samples+i] = q_data[i*sh_band_dim+j];
    }
  }
}

////////////////////////////////////////////////////////////////////////////////
//! Compute kernel matrix for set of sampling \a lambdas
////////////////////////////////////////////////////////////////////////////////
void
Energy::computeKernelMatrix( const LambdaVec& lambdas) {

  // construct kernel matrix
  unsigned int offset = 0;
  for( unsigned int i = 0; i < num_samples; ++i) {

    // evaluate band and copy into kernel matrix
    sh_eval->shEvalBand( sh_band, lambdas[i].first, lambdas[i].second, sh_vals);
    memcpy( q_data + offset, &sh_vals.front(), sizeof(double)*sh_band_dim);

    offset += sh_band_dim;
  }

  // transpose the kernel matrix to match the Fortran format in Lapack
  transposeQ();
}

////////////////////////////////////////////////////////////////////////////////
//! Compute condition number 
//! @return condition number
////////////////////////////////////////////////////////////////////////////////
double
Energy::computeCondNum( const LambdaVec& lambdas) {
  
  // compute kernel matrix for give set of sampling
  computeKernelMatrix( lambdas);

  // compute singular value decomposition
  int info;
  char arg1 = 'N';
  int  arg2 = num_samples;
  int  arg3 = sh_band_dim;
  int  arg4 = 1;
  dgesvd_( &arg1, &arg1, &arg2, &arg3, q_trans, &arg2,
           &svals.front(), NULL, &arg4, NULL, &arg4, 
           ws, &ws_size, &info);
  
  // check return value
  if( 0 != info) {
    std::cerr << "Lapack call failed. info = " << info << std::endl;
    exit(-1);
  }

  // return condition number
  return svals[0] / svals[sh_band_dim-1];
}

