/// @file
////////////////////////////////////////////////////////////////////////////////
///
/// Copyright (C) 2009      University of Toronto
///
////////////////////////////////////////////////////////////////////////////////
///
///  module     : condition number optimization
///
///  author     : lessig@dgp.toronto.edu
///
///  project    : Spherical Harmonics rotation
///
///  description: computation of "energy" for optimization
///
///  license    : See README for details.
///
////////////////////////////////////////////////////////////////////////////////

// includes, file
#include <optimizer.h>

// includes, system
#include <cfloat>
#include <cstdio>
#include <ctime>

// includes, project
#include <file_io.h>

////////////////////////////////////////////////////////////////////////////////
//! Constructor
////////////////////////////////////////////////////////////////////////////////
Optimizer::Optimizer() :
  lambdas(),
  energy_comp(),
  energy( 0),
  gradient()
{ }

////////////////////////////////////////////////////////////////////////////////
//! Destructor
////////////////////////////////////////////////////////////////////////////////
Optimizer::~Optimizer() { }

////////////////////////////////////////////////////////////////////////////////
//! Initialize optimize
//! @param fname_lambdas  file name for the initial points of the optimization
//! @param opt_time  maximum optimization time (in seconds)
////////////////////////////////////////////////////////////////////////////////
void  
Optimizer::init( const std::string& fname_in, const unsigned int opt_time) {

  opt_time_max = opt_time;

  unsigned int sh_band = 0;
  double eps_theta = 0.0;
  double eps_phi = 0.0;

  // read points
  FileIO::readMatlabLambdaFile( fname_in, sh_band, eps_theta, eps_phi, lambdas);

  // initalize energy computation
  energy_comp.init( lambdas.size(), sh_band, eps_theta, eps_phi);

  // adjust temporary storage for gradient
  gradient.resize( 2*lambdas.size());

  // initialize energy and gradient
  energy_comp.compute( lambdas, energy, gradient);
}

////////////////////////////////////////////////////////////////////////////////
//! Run optimizer
////////////////////////////////////////////////////////////////////////////////
void  
Optimizer::run() {

  // init optimizer
  
  unsigned int num_variables = 2*lambdas.size();

  Domain dom( num_variables);
  for( unsigned int i=0; i < num_variables; i++) {
    dom.setBound( i, DBL_MIN, DBL_MAX, Domain::None); 
  }

  double* lambdas_ptr = (double*) &lambdas.front();

  BFGSoptimizer opt( num_variables, lambdas_ptr, &dom, 30);
  opt.setPrintLevel(1); 

  // run optimizer

  time_t start = time(NULL);

  BFGSoptimizer::Results res;
  do { 

    // it is ok to call it first time without computing f and g: 
    // it will return EvaluationRequest  
    res  = opt.nextIteration( energy, &gradient.front());

    if( res == BFGSoptimizer::EvaluationRequest ) { 
      energy_comp.compute( lambdas, energy, gradient);
    } 

    // check if the maximum iteration time has been reached
    time_t curr = time(NULL);
    unsigned int t_delta = (unsigned int)(curr-start);
    std::cout << "band " << energy_comp.getShBand() 
              << " :: t_delta = " <<  t_delta << " seconds.\n"; 
    if( t_delta > opt_time_max) {
      std::cout << "Maximum optimization time reached. Exiting." << std::endl;
      break;
    }

  } while(    (res == BFGSoptimizer::IterationCompleted)
           || (res == BFGSoptimizer::EvaluationRequest));

}

////////////////////////////////////////////////////////////////////////////////
//! Cleanup and write result
//! @param fname_lambdas  file name for optimized points 
////////////////////////////////////////////////////////////////////////////////
void  
Optimizer::finish( const std::string& fname_lambdas) {

  fixDomainLambda();

  FileIO::writeMatlabLambdaFile( fname_lambdas, energy, lambdas);
}

////////////////////////////////////////////////////////////////////////////////
// The optimization is unconstrained but it is desirable that the final points
// are probably limited to [0,\pi] x [0,2\pi]
////////////////////////////////////////////////////////////////////////////////
void 
Optimizer::fixDomainLambda() {

  for( unsigned int i = 0; i < lambdas.size(); ++i) {
    if( lambdas[i].first < 0.0) {
      lambdas[i].first *= -1.0;
      lambdas[i].second += PI;
    }
    else if( lambdas[i].first > PI) {
      lambdas[i].first = 2*PI -  lambdas[i].first;
      lambdas[i].second += PI;
    }
    else if( lambdas[i].first < 0.0) {
      lambdas[i].second += 2*PI;
    }
    else if( lambdas[i].first > 2*PI) {
      lambdas[i].second -= 2*PI;
    }

    DEBUG_ASSERT( (lambdas[i].first >= 0.0) && (lambdas[i].first <= PI),
                  "Optimizer", "fixDomainLambda()", "Invalid range for theta");
    DEBUG_ASSERT( (lambdas[i].second >= 0.0) && (lambdas[i].second <= 2*PI),
                  "Optimizer", "fixDomainLambda()", "Invalid range for theta");

  }
}

