/// @file
////////////////////////////////////////////////////////////////////////////////
///
/// Copyright (C) 2009      University of Toronto
///
////////////////////////////////////////////////////////////////////////////////
///
///  module     : condition number optimization
///
///  author     : lessig@dgp.toronto.edu
///
///  project    : Spherical Harmonics rotation
///
///  description: computation of "energy" for optimization
///
///  license    : See README for details.
///
////////////////////////////////////////////////////////////////////////////////

#ifndef _ENERGY_H_
#define _ENERGY_H_

// includes, Spherical Harmonics
#include <sh_evaluator.h>

//! Compute energy
class Energy {

public:

  // potentials
  static  const unsigned int cn = 1;
  static  const unsigned int fp = 2;

public:

  //! Constructor
  Energy();
  
  //! Destructor
  ~Energy();

public:

  //! Initialization
  void init( const unsigned int num_samples, const unsigned int band, 
             const double eps_theta, const double eps_phi);

  //! Compute energy and gradient at \a lambdas
  template< unsigned int PotentialType >
  void compute( const LambdaVec& lambdas, double& energy, Vec& gradient);

public:

  inline const unsigned int& getShBandMax() const {
    return sh_band_max;
  }

private:

  //! Compute energy wrapper for template branching
  template< unsigned int PotentialType >
  double computeEnergy( const LambdaVec& lambdas);

  //! Transpose kernel matrix to match Fortran format used in Lapack
  void transposeQ();

  //! Compute kernel matrix for set of sampling \a lambdas
  void computeKernelMatrix( const LambdaVec& lambdas);

  //! Compute condition number for current sampling configuration
  //! @return condition number
  double  computeCondNum( const LambdaVec& lambdas);

  //! Compute frame potential for current sampling configuration
  //! @return frame potential 
  double  computeFramePotential( const LambdaVec& lambdas);

private:

  unsigned int sh_band_max;
  unsigned int sh_band_max_dim;
  unsigned int num_samples;

  // optimizer status variables
  double  fd_eps_theta;
  double  fd_eps_phi;

  // singular values, returned for Lapack SVD computation
  Vec  svals;

  // representation of kernel matrix
  double*  q_data;
  double*  q_trans;

  // Lapack workspace size
  double*    ws;
  int   ws_size;

  // Evaluator for SH coefficients
  ShEvaluator<double>*        sh_eval;
  // storage for all SH coefficients up to sh_band_max
  std::vector< std::vector<double> >         sh_vals;

private:

  //! Constructor, copy 
  //! @note Not implemeted
  Energy( Energy&);

  //! Assignment operator
  //! @note Not implemented
  Energy& operator=( Energy&);

};

// Implementation template functions

////////////////////////////////////////////////////////////////////////////////
//! Compute energy
////////////////////////////////////////////////////////////////////////////////
template<unsigned int PotentialType>
double
Energy::computeEnergy( const LambdaVec& lambdas) {

  double energy = 0.0;

  // compute energy
  if( 1 == PotentialType) {
    energy = computeCondNum( lambdas);
  } 
  else if( 2 == PotentialType) {
    energy = computeFramePotential( lambdas);
  }
  else { 
    assert( 0); 
  };

  return energy;
}

////////////////////////////////////////////////////////////////////////////////
//! Compute energy and gradient at \a lambdas
////////////////////////////////////////////////////////////////////////////////
template<unsigned int PotentialType>
void
Energy::compute( const LambdaVec& lambdas, double& energy, Vec& gradient) {

  // compute energy
  energy = computeEnergy<PotentialType>( lambdas);

  // helper variables for computing the gradient
  LambdaVec lambdas_delta_theta( lambdas);
  LambdaVec lambdas_delta_phi( lambdas);
  double energy_delta = 0.0;

  // compute gradient
  for( unsigned int i = 0; i < num_samples; ++i) {

    // gradient theta[i]
    lambdas_delta_theta = lambdas;
    lambdas_delta_theta[i].first += fd_eps_theta;

    energy_delta = computeEnergy<PotentialType>( lambdas_delta_theta);
    gradient[2*i] = (energy_delta - energy) / fd_eps_theta;

    // gradient phi[i]
    lambdas_delta_phi = lambdas;
    lambdas_delta_phi[i].second += fd_eps_phi;

    energy_delta = computeEnergy<PotentialType>( lambdas_delta_phi);
    gradient[2*i+1] = (energy_delta - energy) / fd_eps_phi;
  }
}


#endif // _ENERGY_H_
