/// @file
////////////////////////////////////////////////////////////////////////////////
///
/// Copyright (C) 2009      University of Toronto
///
////////////////////////////////////////////////////////////////////////////////
///
///  module     : condition number optimization
///
///  author     : lessig@dgp.toronto.edu
///
///  project    : Spherical Harmonics rotation
///
///  description: computation of "energy" for optimization
///
///  license    : See README for details.
///
////////////////////////////////////////////////////////////////////////////////

// includes, file
#include <energy.h>

// includes, system
#include <cstring>
#include <cstdio>

// Lapack SVD routine 
extern "C"
int 
dgesvd_(char *jobu, char *jobvt, int *m, int *n, 
	      double *a, int *lda, double *s, double *u, int* ldu,
	      double *vt, int *ldvt, double *work, int *lwork, 
      	int *info);

////////////////////////////////////////////////////////////////////////////////
//! Constructor
////////////////////////////////////////////////////////////////////////////////
Energy::Energy() :
  sh_band_max( 0),
  sh_band_max_dim( 0),
  num_samples( 0),
  fd_eps_theta( 0.0),
  fd_eps_phi( 0.0),
  svals( 0),
  q_data( NULL),
  ws( NULL),
  ws_size( 0),
  sh_eval( ShEvaluator<double>::instance()),
  sh_vals( )
{ }

////////////////////////////////////////////////////////////////////////////////
//! Destructor
////////////////////////////////////////////////////////////////////////////////
Energy::~Energy() {

  // clean up
  if( NULL != q_data) {
    delete[] q_data;
    delete[] q_trans;
  }

  if( NULL != ws) {
    delete[] ws;
  }
}

////////////////////////////////////////////////////////////////////////////////
//! Initialization
////////////////////////////////////////////////////////////////////////////////
void
Energy::init( const unsigned int n_samples, const unsigned int band_max, 
              const double eps_theta, const double eps_phi) 
{
  sh_band_max = band_max;
  sh_band_max_dim = (sh_band_max+1) * (sh_band_max+1);

  num_samples = n_samples;

  fd_eps_theta = eps_theta;
  fd_eps_phi = eps_phi;

  // singular values
  svals.resize( sh_band_max_dim);
  // storage for all SH coefficients up to sh_band_max for one sampling point
  sh_vals.resize( sh_band_max+1);
  for( unsigned int band = 0; band <= sh_band_max; ++band) {
    unsigned int band_dim = 2*band+1;
    sh_vals[band].resize( band_dim);
  }

  q_data = new double[sh_band_max_dim * num_samples];
  q_trans = new double[sh_band_max_dim * num_samples];

  // SH evaluator, has to be initialized to max. band
  sh_eval->init( sh_band_max);

  // Lapack worspace size
  ws_size = 12*num_samples;
  ws = new double[ws_size];
};

////////////////////////////////////////////////////////////////////////////////
//! Transpose kernel matrix to match Fortran format used in Lapack
////////////////////////////////////////////////////////////////////////////////
void 
Energy::transposeQ() {

  for( unsigned int i = 0; i < num_samples; ++i) {
    for( unsigned int j = 0; j < sh_band_max_dim; ++j) {
      q_trans[j*num_samples+i] = q_data[i*sh_band_max_dim+j];
    }
  }
}

////////////////////////////////////////////////////////////////////////////////
//! Compute kernel matrix for set of sampling \a lambdas
////////////////////////////////////////////////////////////////////////////////
void
Energy::computeKernelMatrix( const LambdaVec& lambdas) {

  unsigned int offset = 0;
  // for each samples
  for( unsigned int i = 0; i < num_samples; ++i) {

    // evaluate band 
    sh_eval->shEvalBands( sh_band_max, lambdas[i].first, lambdas[i].second,
                          sh_vals);
    
    // copy coefficients into kernel matrix
    for( unsigned int band = 0; band <= sh_band_max; ++band) {

      // dimension of current band
      unsigned int band_dim = 2*band+1;

      // copy data
      memcpy( q_data + offset, &sh_vals[band].front(),
              sizeof(double) * band_dim);
   
      // update offset
      offset += band_dim;
    }
  }

  // transpose the kernel matrix to match the Fortran format in Lapack
  transposeQ();
}

////////////////////////////////////////////////////////////////////////////////
//! Compute condition number 
//! @return condition number
////////////////////////////////////////////////////////////////////////////////
double
Energy::computeCondNum( const LambdaVec& lambdas) {
  
  // compute kernel matrix for give set of sampling
  computeKernelMatrix( lambdas);

  // compute singular value decomposition
  int info;
  char arg1 = 'N';
  int  arg2 = num_samples;
  int  arg3 = sh_band_max_dim;
  int  arg4 = 1;
  dgesvd_( &arg1, &arg1, &arg2, &arg3, q_trans, &arg2,
           &svals.front(), NULL, &arg4, NULL, &arg4, 
           ws, &ws_size, &info);
  
  // check return value
  if( 0 != info) {
    std::cerr << "Lapack call failed. info = " << info << std::endl;
    exit(-1);
  }

  // return condition number
  // some care is required to take the last non-zero singular value 
  unsigned nzl = std::min( sh_band_max_dim, num_samples);
  return (svals[0] / svals[nzl-1]);
}

////////////////////////////////////////////////////////////////////////////////
//! Compute condition number 
//! @return condition number
////////////////////////////////////////////////////////////////////////////////
double
Energy::computeFramePotential( const LambdaVec& lambdas) {
  
  // compute kernel matrix for give set of sampling
  computeKernelMatrix( lambdas);

  // normalize kernel matrix
  double bm_f = (double) (sh_band_max+1);
  double normalize = 1.0 / sqrt( (bm_f*bm_f) / (4.0*PI));
  for( unsigned int i = 0; i < (num_samples * sh_band_max_dim); ++i) {
    q_trans[i] *= normalize;
  }

  // compute singular value decomposition
  int info;
  char arg1 = 'N';
  int  arg2 = num_samples;
  int  arg3 = sh_band_max_dim;
  int  arg4 = 1;
  dgesvd_( &arg1, &arg1, &arg2, &arg3, q_trans, &arg2,
           &svals.front(), NULL, &arg4, NULL, &arg4, 
           ws, &ws_size, &info);
  
  // check return value
  if( 0 != info) {
    std::cerr << "Lapack call failed. info = " << info << std::endl;
    exit(-1);
  }

  // compute frame potential which is given by the sum of the squared 
  // eigenvalues of the the frame operator, which is the fourth power of 
  // the singular values of the kernel matrix
  double fp = 0.0;
  for( unsigned int i = 0; i < sh_band_max_dim; ++i) {
    fp += svals[i]*svals[i]*svals[i]*svals[i];
  }
 
  // don't use the frame potential directly but use the energy function 
  // proposed by Casazza et al. (2010)
  double ns_f = (double) num_samples;
  fp = sqrt( fp - ((ns_f * ns_f) / ((double) sh_band_max_dim)));

  return fp;
}

