/// @file
////////////////////////////////////////////////////////////////////////////////
///
/// Copyright (C) 2009      University of Toronto
///
////////////////////////////////////////////////////////////////////////////////
///
///  module     : condition number optimization
///
///  author     : lessig@dgp.toronto.edu
///
///  project    : Spherical Harmonics rotation
///
///  description: File I/O for data transfer with matlab
///
///  license    : See README for details.
///
////////////////////////////////////////////////////////////////////////////////

// includes, file
#include <file_io.h>

// includes, system
#include <fstream>
#include <iomanip>

////////////////////////////////////////////////////////////////////////////////
//! Read Matlab ascii file containing the set of sampling { \lambda_i } and
//! return the data in a C compatible format  
//! @param fname  file name for the input file
////////////////////////////////////////////////////////////////////////////////
void
FileIO::
readMatlabLambdaFile( const std::string& path, unsigned int& sh_band_max,
                      unsigned int& potential, 
                      double& eps_theta, double& eps_phi, LambdaVec& lambdas)
{
  std::string fname = path + "/lambda_in.dat";

  std::fstream fs( fname.c_str(), std::fstream::in);
  if( ! fs.good()) {
    std::cerr << "Opening file '" << fname << "' for reading failed. "
              << "Terminating." << std::endl;
    exit(-1);
  }

  // read potential
  double potential_temp;
  fs >> potential_temp;
  potential = (unsigned int) potential_temp;

  // read sh band
  double temp;
  fs >> temp;
  sh_band_max = (unsigned int) temp;

  // read finite difference epsilons
  fs >> eps_theta;
  fs >> eps_phi; 

  // read lambda data
  lambdas.clear();
  double theta, phi = 0.0;
  while( ! fs.eof()) {
    
    // read from file
    fs >> theta;
    fs >> phi;

    // store current lambda
    lambdas.push_back( LambdaType( theta, phi));
  }
  // remove spurious point at the end (due to the loop setup)
  lambdas.pop_back();

  std::cout << "\n==============\n";
  std::cout << "Parameters:\n";
  std::cout << "sh_band_max = " << sh_band_max << '\n';
  std::cout << "eps_theta = " << eps_theta << '\n';
  std::cout << "eps_phi = " << eps_phi << '\n';
  std::cout << "==============\n\n";

  fs.close();
}

////////////////////////////////////////////////////////////////////////////////
//! Write Matlab ascii file
//! @param fname  file name of the output file
////////////////////////////////////////////////////////////////////////////////
void
FileIO::writeMatlabLambdaFile( const std::string& path, const double energy,
                               const LambdaVec& lambdas) 
{
  {
    // open file for writing
    std::string fname = path + "/lambda_out.dat";
    std::fstream fs( fname.c_str(), std::fstream::out);
    if( ! fs.good()) {
      std::cerr << "Opening file '" << fname << "' for writing failed. "
        << "Terminating." << std::endl;
      exit(-1);
    }

    // write data
    fs << std::setprecision( 18);
    for( unsigned int i = 0; i < lambdas.size(); ++i) {
      fs << lambdas[i].first << "  " << lambdas[i].second << "\n";
    }

    // close file
    fs.close();
  }

  {
    // open file for writing
    std::string fname = path + "/energy_out.dat";
    std::fstream fs( fname.c_str(), std::fstream::out);
    if( ! fs.good()) {
      std::cerr << "Opening file '" << fname << "' for writing failed. "
        << "Terminating." << std::endl;
      exit(-1);
    }

    // write data
    fs << std::setprecision( 18);
    fs << energy << "\n";

    // close file
    fs.close();
  }

}

