/// @file
////////////////////////////////////////////////////////////////////////////////
///
/// Copyright (C) 2009      University of Toronto
///
////////////////////////////////////////////////////////////////////////////////
///
///  module     : rotation
///
///  author     : lessig@dgp.toronto.edu
///
///  project    : Spherical Harmonics rotation
///
///  description: Rotation utility functionality
///
///  license    : See README for details.
///
////////////////////////////////////////////////////////////////////////////////

#ifndef _ROTATION_H_
#define _ROTATION_H_

namespace Rotation {

////////////////////////////////////////////////////////////////////////////////
//! Construct rotation matrix for ZYZ Euler angles
//! @param t1 / t2 / t3   Euler angles (in rad)
//! @param R  rotation matrix
////////////////////////////////////////////////////////////////////////////////
template<class T, class MatrixType>
void
getRotationMatrixEulerZYZ(const T& t1, const T& t2, const T& t3, MatrixType& R);

////////////////////////////////////////////////////////////////////////////////
//! Construct rotation matrix for axis, angle
//! @param axis1,axis2,axis3, angle  normalized vector and angle in rad
//! @param R  rotation matrix
////////////////////////////////////////////////////////////////////////////////
template<class T, class MatrixType>
void
getRotationMatrixAxisAngle( const T& x, const T& y, const T& z, const T& angle, 
                            MatrixType& R);

////////////////////////////////////////////////////////////////////////////////
//! Extract Euler ZYZ angles from rotation matrix
//! @param R  rotation matrix
//! @param t1 / t2 / t3   Euler angles (in rad)
////////////////////////////////////////////////////////////////////////////////
template<class T, class MatrixType>
void
getRotationEulerZYZMatrix( const MatrixType& R, T& t1, T& t2, T& t3);

////////////////////////////////////////////////////////////////////////////////
//! Generator for random rotations 
//! @note The implementation is based a memo by Jim Arvo 
//!       (www.ics.uci.edu/~arvo/papers/RotationMat.ps)
////////////////////////////////////////////////////////////////////////////////
template< class T >
class RandomRotationGenerator {

public:

  //! Constructor, default
  RandomRotationGenerator();

  //! Destructor
  ~RandomRotationGenerator();

public:

  // standard interface

  //! Get the next random rotation as Euler ZYZ rotation
  //! @param t1 / t2 / t3  Euler ZYZ rotation angles
  void  get( T& t1, T& t2, T& t3) const;

  //! Get the next random rotation as Euler ZYZ rotation
  //! @note MatrixType has to be a matrix type with mat(i,j) implemented
  //! @param mat  rotation matrix
  template<class MatrixType>
  void  get( MatrixType& mat) const;

private:
  
  //! Constructor, copy
  //! @note Not implemented.
  RandomRotationGenerator( const RandomRotationGenerator&);

  //! Assignment operator
  //! @note Not implemented.
  RandomRotationGenerator& operator=( const RandomRotationGenerator&);

};

} // end namespace Rotation

// includes, implementation
#include <rotation.inl>

#endif // _ROTATION_H_
