/// @file
////////////////////////////////////////////////////////////////////////////////
///
/// Copyright (C) 2009      University of Toronto
///
////////////////////////////////////////////////////////////////////////////////
///
///  module     : sh
///
///  author     : lessig@dgp.toronto.edu
///
///  project    : Spherical Harmonics rotation
///
///  description: Core Spherical Harmonics functionality
///
///  license    : See README for details.
///
////////////////////////////////////////////////////////////////////////////////

#ifndef _SH_EVALUATOR_H_
#define _SH_EVALUATOR_H_

// includes, system
#include <vector>

template< class T >
class ShEvaluator {

  typedef  std::vector<T>                 VecT;
  typedef  std::vector< std::pair<T,T> >  GradVecT;
  typedef  std::vector< VecT >            ShCoeffsVecT;
  typedef  std::vector< GradVecT >        ShGradVecT;

private:

  //! Constructor, default
  ShEvaluator();

  //! Destructor
  ~ShEvaluator();

public:

  // public interface

  //! Initialize
  void  init( const unsigned int max_band);

  //! Evaluate all Spherical Harmonics basis function in the specified band
  //! @param band     Spherical Harmonics band.
  //! @param theta    Spherical coordinate (latitude)
  //! @param phi      Spherical coordinate (longitude)
  //! @param sh_vals  Values of the Spherical Harmonics basis functions (the 
  //!                 linear indexing is from -l to +l)
  void  shEvalBand(const int& band, const T& theta, const T& phi, VecT& sh_vals);

  //! Evaluate all Spherical Harmonics basis functions between band_l and band_u
  //! @param theta  Spherical coordinate (latitude)
  //! @param phi    Spherical coordinate (longitude)
  //! @param sh_vals  Values of the Spherical Harmonics basis functions (the 
  //!                 linear indexing is band after band and from -l to +l
  //!                 within each band)
  void shEvalBands( const int& band_l, const int& band_u, 
                    const T& theta, const T& phi, 
                    ShCoeffsVecT& sh_vals);

  //! Evaluate all Spherical Harmonics basis functions up to and including l
  //! @param max_band   Maximum Spherical Harmonics band.
  //! @param theta  Spherical coordinate (latitude)
  //! @param phi    Spherical coordinate (longitude)
  //! @param sh_vals  Values of the Spherical Harmonics basis functions (the 
  //!                 linear indexing is band after band and from -l to +l
  //!                 within each band)
  void shEvalBands( const int& max_band, const T& theta, const T& phi, 
                    ShCoeffsVecT& sh_vals);

  //! Evaluate all Spherical Harmonics basis functions up to and including l and 
  //! their gradient
  //! @param l      Spherical Harmonics band.
  //! @param theta  Spherical coordinate (latitude)
  //! @param phi    Spherical coordinate (longitude)
  //! @param sh_vals  Values of the Spherical Harmonics basis functions (the 
  //!                 linear indexing is band after band and from -l to +l within
  //!                 each band)
  //! @param sh_vals  Gradient (w.r.t. theta and phi) of sh_vals
  void
  shEvalBands( const int& l, const T& theta, const T& phi, 
               ShCoeffsVecT& sh_vals, ShGradVecT& d_sh_vals );

  // get a handle to the instance
  static  ShEvaluator*  instance();

private:

  //! maximum band for which the instance is initialized
  unsigned int max_band_init;

  //! Precomputed data for longitudinal components
  VecT  sin_table;
  VecT  cos_table;

  // we have to keep track of the values of the Legendre polynomials for the 
  // the last two levels. Allocate the vectors once and for all since the 
  // maximum size is known

  //! Legendre polynomial values
  VecT p_l;
  //! Derivatives of Legendre polynomials
  VecT dp_l;
};

// include implementation
#include <sh_evaluator.inl>

#endif // _SH_EVALUATOR_H_

