/// @file
////////////////////////////////////////////////////////////////////////////////
///
/// Copyright (C) 2009      University of Toronto
///
////////////////////////////////////////////////////////////////////////////////
///
///  module     : sh
///
///  author     : lessig@dgp.toronto.edu
///
///  project    : Spherical Harmonics rotation
///
///  description: Spherical Harmonics rotation using RKHS
///
///  license    : See README for details.
///
////////////////////////////////////////////////////////////////////////////////

#ifndef _SH_ROTATION_RKHS_ZONAL_H_
#define _SH_ROTATION_RKHS_ZONAL_H_

// includes, file
#include <sh_rotation_base.h>

// includes, system
#include <string>
#include <vector>

// includes, libs
#include <cutil.h>
#include <boost/numeric/mtl/mtl.hpp>

// includes, project
#include <point_s2.h>
#include <sh_coeffs.h>

template<class T>
class ShRotationRKHSZonal : public ShRotationBase<T> {

public:

  // MTL matrix and vector convenience typedef's
  typedef  mtl::matrix::dense2D<T>  Matrix;
  typedef  mtl::dense_vector<T>  Vector;

public:

  //! Constructor, default
  //! @param oversampling factor
  //! @param  path   path to data files
  ShRotationRKHSZonal( const unsigned int oversampling_factor = 100,
                  const std::string  path = "");

  //! Destructor
  ~ShRotationRKHSZonal();

public:

  //! Initialize rotation, in particular load set of sampling and change of
  //! basis matrices
  //! @param  max_band  maximum band to pre-load
  //! @param  fname  file name containing the data needed for rotation
  //!                the default name is 'data/sh_rotation_rkhs.dat'
  void preload( const unsigned int max_band);

  //! Load rotation matrices and associated sets of sampling
  //! @param  band  band of matrix to load
  //! @param  fname  file name containing the data needed for rotation
  //!                the default name is 'sh_rotation_rkhs_b*.dat'
  //! @param  cbm_c  matrix will be loaded here
  //! @param  set_sampling_c  sample points will be loaded here
  void loadCbm( const unsigned int band, std::string fname, Matrix& cbm_c,
                       std::vector< PointS2<T> >& set_sampling_c);


  //! Rotate a signal represented in Spherical Harmonics
  //! @return  Computation time for function values (ms).
  //! @param  sh_coeffs  Spherical Harmonics coefficient vector (unrotated)
  //! @param  sh_coeffs_rotated  Spherical Harmonics coefficient vector
  //!                            (after rotation)
  //! @param  max_band   Maximum band of \a sh_coeffs to be rotated (that is
  //!                    \a sh_coeffs has to contain at least max_band bands)
  //! @param  t1 / t2 / t3  Euler ZYZ rotation angles
  //! @param  num_iterations  Number of iterations (for computing rotation
  //!                         matrices and measuring execution time)
  T rotatePreloaded( const ShCoeffs<T>& sh_coeffs,
                     ShCoeffs<T>& sh_coeffs_rotated,
                     const unsigned int max_band,
                     const T& t1, const T& t2, const T& t3,
                     const unsigned int num_iterations = 1);


  //! Rotate a signal represented in Spherical Harmonics. Assumes
  //! matrices are not preloaded.
  //! @return  Computation time for function values (ms).
  //! @param  sh_coeffs  Spherical Harmonics coefficient vector (unrotated)
  //! @param  sh_coeffs_rotated  Spherical Harmonics coefficient vector
  //!                            (after rotation)
  //! @param  max_band   Maximum band of \a sh_coeffs to be rotated (that is
  //!                    \a sh_coeffs has to contain at least max_band bands)
  //! @param  t1 / t2 / t3  Euler ZYZ rotation angles
  //! @param  num_iterations  Number of iterations (for computing rotation
  //!                         matrices and measuring execution time)
  T rotate( const ShCoeffs<T>& sh_coeffs,
            ShCoeffs<T>& sh_coeffs_rotated,
            const unsigned int max_band,
            const T& t1, const T& t2, const T& t3,
            const unsigned int num_iterations);

  //! Get the name of the rotation algorithm
  const std::string& name() const;
  
private:

  // utility functionality

  // calculates the number of samples in a band based on a given oversampling 
  // rate
  unsigned int numSamplesOversampling( unsigned int band,unsigned int os_factor);

  //! Check if the point set in \a set_sampling is nested and set
  //! \a set_sampling_nested appropriately
  void  checkSetOfSamplingNested();

  //! Allocate memory for temporary storage used during rotatePreloaded()
  void  prealloc( const unsigned int os_factor);

private:

  //! Transpose change of basis matrix
  void  
  cbmTranspose( const unsigned int band, const Matrix& cbm_c, Matrix& cbm_c_t);

  //! Evaluate per band signals at rotated sample points for a nested set 
  //! of sampling.
  //! compute FVals using Zonal Harmonics
  //! @param  sh_coeffs_r  Spherical Harmonics coefficient vector (unrotated)
  //! @param  sh_coeffs_rotated  Spherical Harmonics coefficient vector
  //!                            (after rotation)
  //! @param  max_band   Maximum band of \a sh_coeffs to be rotated (that is
  //!                    \a sh_coeffs has to contain at least max_band bands)
  //! @param  R          rotation matrix
  //! @param  iters      Number of iterations (for computing rotation
  //!                    matrices and measuring execution time)
  void computeFValsNestedZonal( const std::vector<Vector>& zh_coeffs,
                                const unsigned int max_band,
                        				const Matrix& R);

private:

  //! Constructor, copy
  //! @note Not implemented
  ShRotationRKHSZonal( const ShRotationRKHSZonal&);

  //! Addignment operator
  //! @note Not implemented
  ShRotationRKHSZonal operator=( const ShRotationRKHSZonal&);

private:

  //! oversampling rate
  const unsigned int os_factor;

  //! Maximal band which has been pre-loaded
  unsigned int num_bands_preloaded;

  //! Flag that indicates if the set of sampling is nested (i.e. the set of
  //! sampling for band m is a subset of the set of sampling for band n with
  //! n > m)
  bool  set_sampling_nested;

  //! Sets of sampling (pre-loaded)
  std::vector< std::vector< PointS2<T> > >  set_sampling;

  //! Change of basis matrices (pre-loaded)
  std::vector< Matrix >  cbm;

  //! Transpose of change of basis matrices for computing Zonal Harmonics
  //! coefficients (pre-loaded)
  std::vector< Matrix >  cbm_t;

  //! Temporary storage for the function values
  std::vector< Vector >  fvals;

  //! path to data files
  std::string  path_data;

private:

  //! Legendre polynomial values generated during computations
  typedef std::vector<T>  Vec;
  std::vector< std::vector<Vec> > leg;

  //! Rotated sampling points
  std::vector< PointS2<T> > rotated_set_sampling;

private:

  // timer for performance measurements
  unsigned int timer;

  std::string class_name;
};

// include implementation
#include <sh_rotation_rkhs_zonal.inl>

#endif

