/// @file
////////////////////////////////////////////////////////////////////////////////
///
/// Copyright (C) 2009      University of Toronto
///
////////////////////////////////////////////////////////////////////////////////
///
///  module     : sh
///
///  author     : {lessig|tyler}@dgp.toronto.edu
///
///  project    : Spherical Harmonics rotation
///
///  description: 
///
///  license    : See README for details. 
///
////////////////////////////////////////////////////////////////////////////////

#ifndef _SH_ROTATION_SRKHS_H_
#define _SH_ROTATION_SRKHS_H_

// includes, files
#include <sh_rotation_base.h>

// includes, system
#include <string>

// includes, libs
#include <cutil.h>

// includes, project
#include <sh_coeffs.h>

//! Spherical Harmonics rotation as proposed by Greengard and Gimbutas
template<class T>
class ShRotationSRKHS : public ShRotationBase<T> {

  // MTL matrix and vector convenience typedef's
  typedef  mtl::matrix::dense2D<FType>  Matrix;
  typedef  mtl::dense_vector<FType>  Vector;
  typedef  std::pair<T,T>  ShGrad;

public:

  //! Constructor, default
  ShRotationSRKHS();

  //! Destructor
  ~ShRotationSRKHS();

public:

  //! Initialize rotation, in particular load precomputed data
  //! @param  max_band  maximum band to pre-load
  void preload( const unsigned int max_band);

  //! Rotate a signal represented in Spherical Harmonics
  //! @return  Computation time of rotation matrix (ms).
  //! @param  sh_coeffs  Spherical Harmonics coefficient vector (unrotated)
  //! @param  sh_coeffs_rotated  Spherical Harmonics coefficient vector 
  //!                            (after rotation)
  //! @param  max_band   Maximum band of \a sh_coeffs to be rotated (that is 
  //!                    \a sh_coeffs has to contain at least max_band bands)
  //! @param  t1 / t2 / t3   ZYZ Euler rotation angles.
  //! @param  num_iterations  Number of iterations (for computing rotation
  //!                         matrices and measuring execution time)
  T rotatePreloaded( const ShCoeffs<T>& sh_coeffs, 
                     ShCoeffs<T>& sh_coeffs_rotated, 
                     const unsigned int max_band,
                     const T& t1, const T& t2, const T& t3,
                     const unsigned int num_iterations = 1);

  //! Rotate a signal represented in Spherical Harmonics
  //! @return  Computation time of rotation matrix (ms).
  //! @param  sh_coeffs  Spherical Harmonics coefficient vector (unrotated)
  //! @param  sh_coeffs_rotated  Spherical Harmonics coefficient vector 
  //!                            (after rotation)
  //! @param  max_band   Maximum band of \a sh_coeffs to be rotated (that is 
  //!                    \a sh_coeffs has to contain at least max_band bands)
  //! @param  t1 / t2 / t3   ZYZ Euler rotation angles.
  //! @param  num_iterations  Number of iterations (for computing rotation
  //!                         matrices and measuring execution time)
  T rotate( const ShCoeffs<T>& sh_coeffs, 
            ShCoeffs<T>& sh_coeffs_rotated, 
            const unsigned int max_band,
            const T& t1, const T& t2, const T& t3,
            const unsigned int num_iterations = 1);
  
  //! Get the name of the rotation algorithm
  const std::string& name() const;
 
private:

  //! Compute sample values for the function and its derivative
  //! @param max_band  maximum band
  //! @param pts       sample points (rotated)
  //! @param beta      rotation angle (a 
  void  computeSampleValues( ShCoeffs<T>& sh_coeffs,
                             const unsigned int max_band, 
                             std::vector< PointS2<T> >& pts);

private:

  //! Constructor, copy
  //! @note Not implemented
  ShRotationSRKHS( const ShRotationSRKHS&);

  //! Addignment operator
  //! @note Not implemented
  ShRotationSRKHS operator=( const ShRotationSRKHS&);

private:

  //! latitute of sampling points
  const T  zeta;
  const T  cos_zeta;
  const T  sin_zeta;

  //! Temporary storage for the function values
  std::vector< std::vector<T> >  f_vals;

  //! Pre-compute scaled associated Legendre polynomial values for mapping 
  //! Fourier coefficients to Spherical Harmonic coefficients
  std::vector< std::vector<T> >  p_nm_scaled;

  //! Timer for performance measurements
  unsigned int timer;

  static const std::string class_name;

};

// include implementation
#include <sh_rotation_srkhs.inl>

#endif 

