/// @file
////////////////////////////////////////////////////////////////////////////////
///
/// Copyright (C) 2008      University of Toronto
///
////////////////////////////////////////////////////////////////////////////////
///
///  module     : util
///
///  author     : lessig@dgp.toronto.edu
///
///  project    : Spherical Harmonics rotation
///
///  description: Miscellaneous functionality.
///
///  license    : See README for details.
///
////////////////////////////////////////////////////////////////////////////////

#ifndef _UTIL_H_
#define _UTIL_H_

// includes, system
#include <cstdlib>
#include <cstdio>
#include <cmath>
#include <cmath>
#include <vector>
#ifndef _WIN32
# include <string.h>
#endif
#ifdef _WIN32
# include <windows.h>
#endif

#ifdef _WIN32
# undef min
# undef max
#endif 

////////////////////////////////////////////////////////////////////////////////
//! Convert degree to radians
////////////////////////////////////////////////////////////////////////////////
template<class T>
T
deg_to_rad( const T& deg) {
  return (deg * PI) / 180.0;
}

////////////////////////////////////////////////////////////////////////////////
//! Compute dot product between two vectors of length \a n
////////////////////////////////////////////////////////////////////////////////
template<class T, class S>
inline T
dot( const T* lhs, const S* rhs, const unsigned int n) {

  T rval = lhs[0] * rhs[0];
  for( unsigned int i = 1; i < n; ++i) {
    rval += lhs[i] * rhs[i];
  }

  return rval;
}

////////////////////////////////////////////////////////////////////////////////
//! Safely free() for pointer
////////////////////////////////////////////////////////////////////////////////
template<class T>
inline void
freePtr( T* &ptr) {

  if( NULL != ptr) {
    free( ptr);
    ptr = NULL;
  }
}

////////////////////////////////////////////////////////////////////////////////
//! Minimum
////////////////////////////////////////////////////////////////////////////////
template<class T>
#ifdef __CUDACC__
__host__  __device__
#endif 
T
min( const T& lhs, const T& rhs) {

  return (lhs < rhs) ? lhs : rhs;
}

////////////////////////////////////////////////////////////////////////////////
//! Maximum
////////////////////////////////////////////////////////////////////////////////
template<class T>
#ifdef __CUDACC__
__host__  __device__
#endif
T
max( const T& lhs, const T& rhs) {

  return (lhs < rhs) ? rhs : lhs;
}

////////////////////////////////////////////////////////////////////////////////
//! Check if \a i is in [\a left,\a right)
//! @return 1  if true, otherwise 0
////////////////////////////////////////////////////////////////////////////////
template<class T>
#ifdef __CUDACC__
__device__
#endif
int
in_interval( const T& i, const T& left, const T& right) {

  return ((i >= left) && (i < right));
}

////////////////////////////////////////////////////////////////////////////////
//! Sign of number (integer data type)
////////////////////////////////////////////////////////////////////////////////
template<class T>
#ifdef __CUDACC__
__host__  __device__
#endif
T
sign_i( const T& val) {
  return (val < 0) ? -1 : 1; 
}

////////////////////////////////////////////////////////////////////////////////
//! Sign of number (float)
////////////////////////////////////////////////////////////////////////////////
#ifdef __CUDACC__
__host__  __device__
#endif
inline float
sign_f( const float& val) {
  return (val < 0.0f) ? -1.0f : 1.0f; 
}

////////////////////////////////////////////////////////////////////////////////
//! Sign of number (double)
////////////////////////////////////////////////////////////////////////////////
#ifdef __CUDACC__
__host__  __device__
#endif
inline double
sign_d( const double& val) {
  return (val < 0.0) ? -1.0 : 1.0; 
}

////////////////////////////////////////////////////////////////////////////////
//! Swap \a lhs and \a rhs
////////////////////////////////////////////////////////////////////////////////
template<class T>
#ifdef __CUDACC__
__host__  __device__
#endif
inline void
swap( T& lhs, T& rhs) {

  T temp = rhs;
  rhs = lhs;
  lhs = temp;
}

///////////////////////////////////////////////////////////////////////////////
//! Normalized random number 
///////////////////////////////////////////////////////////////////////////////
template<class T>
T
getRandNormal() {

  return ((T) rand()) / ((T) RAND_MAX); 
}

///////////////////////////////////////////////////////////////////////////////
//! Print bit pattern
///////////////////////////////////////////////////////////////////////////////
template<class T>
void printBits32( T val) {

  unsigned int* temp = (unsigned int*) &val;
  unsigned int  ival = *temp;

  int mask = 1;
  for( int i = 31; i >= 0; --i) {
    printf( "%i", (int) ((ival >> i) & mask));;
  }
  printf("\n");
}

////////////////////////////////////////////////////////////////////////////////
//! Compute midpoint of interval [\a left, \a right] avoiding overflow
//! @param left   left / lower limit of interval
//! @param right  right / upper limit of interval
////////////////////////////////////////////////////////////////////////////////
#ifdef __CUDACC__
__host__ __device__ 
#endif
inline float 
computeMidpoint( const float left, const float right) {

  float mid;
  if( sign_f( left) == sign_f( right)) {
    mid = left + (right - left) * 0.5f;
  }
  else {
    mid = (left + right) * 0.5f;
  }

  return mid;
}

////////////////////////////////////////////////////////////////////////////////
//! Compute weighted midpoint of interval [\a left, \a right] avoiding 
//! overflow 
//! @param left   left / lower limit of interval
//! @param right  right / upper limit of interval
////////////////////////////////////////////////////////////////////////////////
#ifdef __CUDACC__
__host__ __device__ 
#endif
inline float 
computeWeightedMidpoint( const float left, const float right, const float w) 
{
  return left + (right - left) * w;
}

////////////////////////////////////////////////////////////////////////////////
//! Compute the next lower power of two of n
//! @param  n  number for which next higher power of two is seeked
////////////////////////////////////////////////////////////////////////////////
#ifdef __CUDACC__
__host__ __device__ 
#endif
inline int
floorPow2(int n) {

  // early out if already power of two
  if( 0 == (n & (n-1))) {
    return n;
  }

  int exp;
  frexp((float)n, &exp);
  return (1 << (exp - 1));
}

////////////////////////////////////////////////////////////////////////////////
//! Compute the next higher power of two of n
//! @param  n  number for which next higher power of two is seeked
////////////////////////////////////////////////////////////////////////////////
#ifdef __CUDACC__
__host__ __device__ 
#endif
inline int
ceilPow2(int n) {

  // early out if already power of two
  if( 0 == (n & (n-1))) {
    return n;
  }

  int exp;
  frexp((float)n, &exp);
  return (1 << exp);
}

////////////////////////////////////////////////////////////////////////////////
//! Check for CUDA errors 
//! @note In contrast to th existing macro in CUTIL this one is always active
//!       independent of the build type (debug or release)
////////////////////////////////////////////////////////////////////////////////
#define CHECK_CUDA_ERROR( msg)                                              \
  cudaError_t err = cudaGetLastError();                                     \
  if( cudaSuccess != err) {                                                 \
    fprintf(stderr, "Cuda error: %s in file '%s' in line %i : %s.\n",       \
            msg, __FILE__, __LINE__, cudaGetErrorString( err) );            \
    exit(EXIT_FAILURE);                                                     \
  }                                                                         \

////////////////////////////////////////////////////////////////////////////////
//! Highly accurate timing function
//! @return clock cycles since system boot
////////////////////////////////////////////////////////////////////////////////
#ifdef _WIN32
inline __declspec(naked)
unsigned __int64 __cdecl rdtsc() {
   __asm
   {
      rdtsc
      ret       ; return value at EDX:EAX
   }
}
#else
extern __inline__ 
unsigned long long int rdtsc() {

  unsigned long long int x;
  __asm__ volatile (".byte 0x0f, 0x31" : "=A" (x));
  return x;
}
#endif

////////////////////////////////////////////////////////////////////////////////
//! Determine CPU clock speed (helper for rdtsc)
//! @note 1 / f, where f is the returned value, are the micro second (10^-6 
//!       seconds) for one cycle
////////////////////////////////////////////////////////////////////////////////
extern "C" {

#ifdef _WIN32
inline
void GetTSC(unsigned long& tsc)
{
  unsigned long a = 0;
  __asm _emit 0fh
  __asm _emit 31h
  __asm mov a, eax;
  tsc=a;
}
#endif

inline float
getCPUClockSpeed() {

#ifdef _WIN32

  LARGE_INTEGER    s, e, freq;
  unsigned long cs, ce;

  // Determine timer frequency.
  QueryPerformanceFrequency(&freq);

  QueryPerformanceCounter(&s);
  GetTSC( cs);

  // Wait for a while...
  for (volatile long i = 0; i < 1000000; ++i) ;

  GetTSC( ce);
  QueryPerformanceCounter(&e);
 
  // Calculate frequency.
  return (float) ((ce - cs) * freq.QuadPart / (e.QuadPart - s.QuadPart));

#else // LINUX

  FILE *fp;
  char buffer[4096];
  size_t bytes_read;
  char *match;
  float clock_speed;

  fp = fopen ("/proc/cpuinfo", "r");
  bytes_read = fread (buffer, 1, sizeof (buffer), fp);
  fclose (fp);
  if (bytes_read == 0 || bytes_read == sizeof (buffer)) {
   return 0;
  }
  buffer[bytes_read] = '\0';
  match = strstr (buffer, "cpu MHz");
  if (match == NULL)
  return 0;
  sscanf (match, "cpu MHz : %f", &clock_speed);
  return clock_speed;

#endif
}
} // end extern "C"


#endif // #ifndef _UTIL_H_
