/// @file
////////////////////////////////////////////////////////////////////////////////
///
/// Copyright (C) 2009      University of Toronto
///
////////////////////////////////////////////////////////////////////////////////
///
///  module     : sh
///
///  author     : lessig@dgp.toronto.edu
///
///  project    : Spherical Harmonics rotation
///
///  description: Point on the sphere S^2
///
///  license    : See README for details.
///
////////////////////////////////////////////////////////////////////////////////

// includes, project
#include <rotation.h>

////////////////////////////////////////////////////////////////////////////////
//! Constructor, default (creates a point on the North pole)
////////////////////////////////////////////////////////////////////////////////
template<class T>  
PointS2<T>::PointS2() :
  p_x(0.0), p_y( 0.0), p_z(0.0)
{ }

////////////////////////////////////////////////////////////////////////////////
//! Constructor
//! @param  t1 / t2 / t3  x / y / z location of point 
////////////////////////////////////////////////////////////////////////////////
template<class T>
PointS2<T>::PointS2( const T& t1, const T& t2, const T& t3) :
  p_x( t1), p_y( t2), p_z( t3)
{ }

////////////////////////////////////////////////////////////////////////////////
//! Constructor
//! @param  theta / phi  location of point in spherical coordinates
////////////////////////////////////////////////////////////////////////////////
template<class T>
PointS2<T>::PointS2( const T& theta, const T& phi) :
  p_x( sin(theta) * cos(phi)),
  p_y( sin(theta) * sin(phi)),
  p_z( cos(theta))
{
  DEBUG_ASSERT( ((theta >= 0.0) && (theta <= PI)), "PointS2", "PointS2()", 
                "Invalid range for theta");
  DEBUG_ASSERT( ((phi >= 0.0) && (phi <= 2*PI)), "PointS2", "PointS2()", 
                "Invalid range for phi");
}
////////////////////////////////////////////////////////////////////////////////
//! Destructor
////////////////////////////////////////////////////////////////////////////////
template<class T>
PointS2<T>::~PointS2() { }

////////////////////////////////////////////////////////////////////////////////
//! Constructor, copy
////////////////////////////////////////////////////////////////////////////////
template<class T>
PointS2<T>::PointS2( const PointS2& p) :
  p_x( p.p_x), p_y( p.p_y), p_z( p.p_z)
{ }

////////////////////////////////////////////////////////////////////////////////
//! Assignment operator
////////////////////////////////////////////////////////////////////////////////
template<class T>
PointS2<T>
PointS2<T>::operator=( const PointS2& p) {

  if( this != &p) {
    p_x = p.p_x;
    p_y = p.p_y;
    p_z = p.p_z;
  }

  return p;
}

////////////////////////////////////////////////////////////////////////////////
//! Access x component
////////////////////////////////////////////////////////////////////////////////
template<class T>
T&
PointS2<T>::x() {
  return p_x;
}

////////////////////////////////////////////////////////////////////////////////
//! Access x component
////////////////////////////////////////////////////////////////////////////////
template<class T>
const T&
PointS2<T>::x() const {
  return p_x;
}

////////////////////////////////////////////////////////////////////////////////
//! Access y component
////////////////////////////////////////////////////////////////////////////////
template<class T>
T&
PointS2<T>::y() { 
  return p_y;
}

////////////////////////////////////////////////////////////////////////////////
//! Access y component
////////////////////////////////////////////////////////////////////////////////
template<class T>
const T&
PointS2<T>::y() const {
  return p_y;
}

////////////////////////////////////////////////////////////////////////////////
//! Access z component
////////////////////////////////////////////////////////////////////////////////
template<class T>
T&
PointS2<T>::z() {
  return p_z;
}

////////////////////////////////////////////////////////////////////////////////
//! Access z component
////////////////////////////////////////////////////////////////////////////////
template<class T>
const T&
PointS2<T>::z() const {
  return p_z;
}

////////////////////////////////////////////////////////////////////////////////
//! Access theta
////////////////////////////////////////////////////////////////////////////////
template<class T>
const T 
PointS2<T>::theta() const {

#ifdef DEBUG
  // check that the vector is normalized
  assert( std::abs(1 - ((p_x*p_x) + (p_y*p_y) + (p_z*p_z))) < EPSILON);
#endif

  return acos(p_z);
}

////////////////////////////////////////////////////////////////////////////////
//! Access phi
////////////////////////////////////////////////////////////////////////////////
template<class T>
const T 
PointS2<T>::phi() const {
 
#ifdef DEBUG
  // check that the vector is normalized
  assert( std::abs(1 - ((p_x*p_x) + (p_y*p_y) + (p_z*p_z))) < EPSILON);
#endif

  T theta = acos(p_z);
  T phi = 0.0;

  // early out if at a pole
  if( (0 == theta) || (PI == theta)) {
    return phi;
  }

  // first handle special cases that on equator and either x or y are 0
  // and then general situation
  if( 0 == p_x) {
    if( p_y > 0) {
      phi = PI/2;
    }
    else {
      phi = 3.0/2.0 * PI;
    }
  }
  else if( 0 == p_y) {
    if( p_x > 0) {
      phi = 0.0;
    }
    else {
      phi = PI;
    }
  }
  else { 
    if( (p_x > 0) && (p_y > 0)) {
      phi = atan( p_y / p_x);
    }
    else if( (p_x < 0) && (p_y > 0)) {
      phi = PI - atan( p_y / std::abs(p_x));
    }
    else if( (p_x < 0) && (p_y < 0)) {
      phi = PI + atan( std::abs(p_y) / std::abs(p_x));
    }
    else if( (p_x > 0) && (p_y < 0)) {
      phi = 2 * PI - atan( std::abs(p_y) / p_x);
    }
    else {
      exit(0);
    }
  }

  return phi;
}


////////////////////////////////////////////////////////////////////////////////
//! Rotate point using Euler ZYZ angles
//! @param z / y / z  Euler ZYZ angles
////////////////////////////////////////////////////////////////////////////////
template<class T>
template<class S>
void 
PointS2<T>::rotate( const S& z1, const S& y, const S& z2) {

  mtl::matrix::dense2D<T>  R(3,3);
  Rotation::getRotationMatrixEulerZYZ( z1, y, z2, R);

  rotate( R);
}

////////////////////////////////////////////////////////////////////////////////
//! Rotate point 
//! @param R  rotation matrix
////////////////////////////////////////////////////////////////////////////////
template<class T>
template<class MatrixType>
void 
PointS2<T>::rotate( const MatrixType& R) {

  T  p_x_old = p_x;
  T  p_y_old = p_y;
  T  p_z_old = p_z;

  p_x = R(0,0) * p_x_old + R(0,1) * p_y_old + R(0,2) * p_z_old;
  p_y = R(1,0) * p_x_old + R(1,1) * p_y_old + R(1,2) * p_z_old;
  p_z = R(2,0) * p_x_old + R(2,1) * p_y_old + R(2,2) * p_z_old;
}



////////////////////////////////////////////////////////////////////////////////
//! Dot Product 
//! @param R  rotation matrix
////////////////////////////////////////////////////////////////////////////////
template<class T>
T
PointS2<T>::dot( const PointS2<T>& p2) {
  return p2.p_x * p_x + p2.p_y* p_y + p2.p_z * p_z;
}

////////////////////////////////////////////////////////////////////////////////
//! Equality operator
////////////////////////////////////////////////////////////////////////////////
template<class T>
bool 
PointS2<T>::operator==( const PointS2<T>& rhs) const {

  return ((p_x == rhs.p_x) && (p_z == rhs.p_z) && (p_z == rhs.p_z));
}

////////////////////////////////////////////////////////////////////////////////
//! Inequality operator
////////////////////////////////////////////////////////////////////////////////
template<class T>
bool 
PointS2<T>::operator!=( const PointS2<T>& rhs) const {
  return (! (*this == rhs));
}


