/// @file
////////////////////////////////////////////////////////////////////////////////
///
/// Copyright (C) 2009      University of Toronto
///
////////////////////////////////////////////////////////////////////////////////
///
///  module     : sh
///
///  author     : lessig@dgp.toronto.edu
///
///  project    : Spherical Harmonics rotation
///
///  description: Spherical Harmonics rotation using the algorithm (and code)
///               by Joe Ivanic.
///
///  license    : See README for details. The rotation code used for this method
///               has been obtained from Leslie Greengard.
///
////////////////////////////////////////////////////////////////////////////////

// includes, system
#include <vector>
#include <iostream>

// includes, boost
#include <boost/numeric/mtl/mtl.hpp>

// includes, project
#include <rotation.h>

// namespace, unnamed
namespace {

  // MTL matrix and vector convenience typedef's
  typedef  mtl::matrix::dense2D<FType>  Matrix;
  typedef  mtl::dense_vector<FType>  Vector;

  //////////////////////////////////////////////////////////////////////////////
  //! Resize Fortran compatible representation of SH rotation matrix
  //////////////////////////////////////////////////////////////////////////////
  template<class T>
  void
  resizeFortranArray( const unsigned int& max_band,    
                      std::vector<T>& sh_fortran_array)
  {
    unsigned int l = max_band;
    sh_fortran_array.resize( (l*(l+1)*(l+2)*4/3)+(l+1));
  }
 
  //////////////////////////////////////////////////////////////////////////////
  //! Get rotation matrix in Fortran format
  //////////////////////////////////////////////////////////////////////////////
  template<class T>
  void
  getRotationMatrixFortran( const T& t1, const T& t2, const T& t3,
                            std::vector<T>& rot_matrix )
  {
    // 3 x 3 matrix
    rot_matrix.resize(9);

    Matrix temp( 3,3);
    Rotation::getRotationMatrixEulerZYZ( t1, t2, t3, temp);

    // interchange rows and columns to match Fortran
    for( unsigned int i = 0; i < 3; ++i) {
      for( unsigned int j = 0; j < 3; ++j) {
        unsigned int index = 3*j + i;
        rot_matrix[index] = temp(i,j);
      }
    }
  }

  //////////////////////////////////////////////////////////////////////////////
  //! Helper function for createShRotationMatrixBand()
  //////////////////////////////////////////////////////////////////////////////
  void
  mapIndicesIvanic( const unsigned int index, unsigned int& m_index,
                    unsigned int& pos, unsigned int& neg, 
                    const unsigned int band_size) 
  {
      if( 0 == index) {
        m_index = band_size/2;
      }
      else if( 0 == (index % 2)) {
        m_index = neg;
        --neg;
      }
      else {
        m_index = pos;
        ++pos;
      }

  }

  //////////////////////////////////////////////////////////////////////////////
  //! Create rotation matrix mat for one band
  //////////////////////////////////////////////////////////////////////////////
  template<class M, class T>
  void
  createShRotationMatrixBand( M& mat, const unsigned int& band_size,  
                              const std::vector<T>& sh_rot_matrix, 
                              const unsigned int band )
  {
    unsigned int l = band-1;
    unsigned int offset = (l*(l+1)*(l+2)*4/3)+(l+1);

    // construct rotation matrix for current band
    const T* ptr = &(sh_rot_matrix.front()) + offset;
    unsigned int r_pos = band_size/2+1;
    unsigned int r_neg = band_size/2-1;
    for( unsigned int i = 0; i < band_size; ++i) {

      // map from ivanic to our indexing scheme  
      unsigned int row = 0;
      mapIndicesIvanic( i, row, r_pos, r_neg, band_size);

      unsigned int c_pos = band_size/2+1;
      unsigned int c_neg = band_size/2-1;

      for( unsigned int j = 0; j < band_size; ++j) {

        // map from ivanic to our indexing scheme  
        unsigned int col = 0;
        mapIndicesIvanic( j, col, c_pos, c_neg, band_size);

        mat(row,col) = ptr[j*band_size+i];
      }
    }
  }

} // end namespace, unnamed

extern "C" {

  //////////////////////////////////////////////////////////////////////////////
  //! Compute SH rotation matrix using the algorithm (and the code) by Ivanic 
  //! @param  A  SH rotation matrix (flattened as a 1D array)
  //! @param  max_band  maximum band for rotation
  //! @param  R     rotation matrix (flattened as 1D array of size 9)
  //! @param  work  scratch array of size 2*max_band
  //! @note   Function is implemented in Fortran
  //////////////////////////////////////////////////////////////////////////////
  void rotmat_( double* A, const unsigned int* l, double* R, double* work);

} // end extern "C"

/*static*/ 
template<class T>
const std::string ShRotationIvanic<T>::class_name = "ShRotationIvanic";

////////////////////////////////////////////////////////////////////////////////
//! Constructor, default
////////////////////////////////////////////////////////////////////////////////
template<class T>
ShRotationIvanic<T>::ShRotationIvanic() :
  ShRotationBase<T>(),
  timer( 0)
{
  cutCreateTimer( &timer);
}

////////////////////////////////////////////////////////////////////////////////
//! Destructor
////////////////////////////////////////////////////////////////////////////////
template<class T>
ShRotationIvanic<T>::~ShRotationIvanic() { 

  cutDeleteTimer( timer);
}


////////////////////////////////////////////////////////////////////////////////
//! Initialize rotation, in particular load precomputed data
//! @param  max_band  maximum band to pre-load
//! @param  fname  file name containing the data needed for rotation
//!                the default name is 'data/pinchonNumMatJ.dat'
////////////////////////////////////////////////////////////////////////////////
template<class T>
void 
ShRotationIvanic<T>::preload( const unsigned int /*max_band*/) { }

////////////////////////////////////////////////////////////////////////////////
//! Rotate a signal represented in Spherical Harmonics
//! @return  Computation time of rotation matrix (ms).
//! @param  sh_coeffs  Spherical Harmonics coefficient vector (unrotated)
//! @param  sh_coeffs_rotated  Spherical Harmonics coefficient vector 
//!                            (after rotation)
//! @param  max_band   Maximum band of \a sh_coeffs to be rotated (that is 
//!                    \a sh_coeffs has to contain at least max_band bands)
//! @param  t1 / t2 / t3   ZYZ Euler rotation angles.
//! @param  num_iterations  Number of iterations (for computing rotation
//!                         matrices and measuring execution time)
////////////////////////////////////////////////////////////////////////////////
template<class T>
T
ShRotationIvanic<T>::rotate( const ShCoeffs<T>& sh_coeffs, 
                             ShCoeffs<T>& sh_coeffs_rotated, 
                             const unsigned int max_band,
                             const T& t1, const T& t2, const T& t3,
                             const unsigned int num_iterations)
{
  // allocate memory for rotation matrix
  std::vector<T> sh_rot_matrix;
  resizeFortranArray( max_band, sh_rot_matrix);

  // get rotation matrix from Euler angles
  std::vector<T> rot_matrix;
  getRotationMatrixFortran( t1, t2, t3, rot_matrix);

  // temporary working memory
  std::vector< double > work( 2*max_band+2, 0.0);
  
  // start timing
  cutStartTimer( timer);

  // perform rotation
  // the offset of +2 in the work array is necessary due to a problem in the
  // ivanic code
  for( unsigned int i = 0; i < num_iterations; ++i) {

    // compute rotation matrix
    rotmat_( &(sh_rot_matrix.front()), &max_band, 
             &(rot_matrix.front()), &(work.front())+2);

    // perform rotation by applying rotation matrices

    // helper variable 
    std::vector<T>  temp( 2*max_band+1, 0.0);

    // apply rotation matrix band by band

    // first band is invariant under rotation
    sh_coeffs_rotated(0,0) = sh_coeffs(0,0);

    for( unsigned int band = 1; band <= max_band; ++band) {

      const unsigned int band_size = 2*band+1;

      Matrix cbm_band( band_size, band_size);
      createShRotationMatrixBand( cbm_band, band_size, sh_rot_matrix, band);

      // construct vector of SH coefficients in the current band
      // first create temporary
      memcpy( &(temp.front()), sh_coeffs.getAddress( band),sizeof(T)*band_size);
      Vector coeffs( band_size, &(temp.front()));

      // create vector for result
      Vector coeffs_rotated( 2*band+1, sh_coeffs_rotated.getAddress(band));

      // perform matrix vector multiplication
      mult( cbm_band, coeffs, coeffs_rotated);
    }
  }

  // stop timer
  cutStopTimer( timer);

  // get timer value and then reset for next call
  T ret_val = cutGetTimerValue( timer) / ((T) num_iterations);
  cutResetTimer( timer);

  return ret_val;
}


////////////////////////////////////////////////////////////////////////////////
//! Rotate a signal represented in Spherical Harmonics
//! @return  Computation time of rotation matrix (ms).
//! @param  sh_coeffs  Spherical Harmonics coefficient vector (unrotated)
//! @param  sh_coeffs_rotated  Spherical Harmonics coefficient vector 
//!                            (after rotation)
//! @param  max_band   Maximum band of \a sh_coeffs to be rotated (that is 
//!                    \a sh_coeffs has to contain at least max_band bands)
//! @param  t1 / t2 / t3   ZYZ Euler rotation angles.
//! @param  num_iterations  Number of iterations (for computing rotation
//!                         matrices and measuring execution time)
////////////////////////////////////////////////////////////////////////////////
template<class T>
T
ShRotationIvanic<T>::rotatePreloaded( const ShCoeffs<T>& sh_coeffs, 
                                         ShCoeffs<T>& sh_coeffs_rotated, 
                                         const unsigned int max_band,
                                         const T& t1, const T& t2, const T& t3,
                                         const unsigned int num_iterations)
{
  //std::cout << "WARNING :: ShRotationIvanic<T>::rotatePreloaded() : "
  //          << "No precomputation implemented." << std::endl; 

  return rotate( sh_coeffs, sh_coeffs_rotated, max_band,
                 t1, t2, t3, num_iterations); 
}

////////////////////////////////////////////////////////////////////////////////
//! Get the name of the rotation algorithm
////////////////////////////////////////////////////////////////////////////////
template<class T>
const std::string& 
ShRotationIvanic<T>::name() const {
  return class_name;
}
  
