/// @file
////////////////////////////////////////////////////////////////////////////////
///
/// Copyright (C) 2009      University of Toronto
///
////////////////////////////////////////////////////////////////////////////////
///
///  module     : sh
///
///  author     : lessig@dgp.toronto.edu
///
///  project    : Spherical Harmonics rotation
///
///  description: Spherical Harmonics rotation using the algorithm (and the
///               implementation of) Greengard and Gimbutas.
///
///  license    : See README for details. The rotation code used for this method
///               has been obtained from Leslie Greengard.
///
////////////////////////////////////////////////////////////////////////////////

// includes, system
#include <vector>
#include <iostream>

// namespace, unnamed
namespace {

  //////////////////////////////////////////////////////////////////////////////
  //! Representation for complex number
  //////////////////////////////////////////////////////////////////////////////
  template<class T>  
  struct Complex {

    Complex() :
      r(0.0),
      i(0.0)
    { };

    template<class S>
    friend std::ostream& operator<<( std::ostream &stream, const Complex<S>& c);

    T r; 
    T i;
  };

  //////////////////////////////////////////////////////////////////////////////
  // Array of complex numbers
  //////////////////////////////////////////////////////////////////////////////
  typedef  std::vector< Complex<double> >  VecCd;
  typedef  std::vector< Complex<float> >   VecCf;

  //////////////////////////////////////////////////////////////////////////////
  //! Stream output for complex numbers
  //////////////////////////////////////////////////////////////////////////////
  template<class S>
  std::ostream& 
  operator<<( std::ostream &stream, const Complex<S>& c) {

    stream << c.r << " + " << c.i << "i" << "  "; 
    return stream;
  }

  //////////////////////////////////////////////////////////////////////////////
  //! Resize Fortran compatible representation of SH coefficients
  //////////////////////////////////////////////////////////////////////////////
  template<class T>
  void
  resizeFortranArray( const unsigned int max_band,    
                      std::vector< Complex<T> >& sh_fortran_array)
  {
    sh_fortran_array.resize( (max_band+1) * (2*(max_band+1)+1));
  }
  

  //////////////////////////////////////////////////////////////////////////////
  //! Create Fortran compatible representation of SH coefficient vector
  //////////////////////////////////////////////////////////////////////////////
  template<class T>
  void
  convertShToFortranArray( const unsigned int max_band,    
                         const ShCoeffs<T>& sh_coeffs, 
                         std::vector< Complex<T> >& sh_fortran_array)
  {
    // resize the array
    resizeFortranArray( max_band, sh_fortran_array);

    // swap content (respecting row / column interchange)
    for( unsigned int band = 0; band < max_band+1; ++band) {
      unsigned int offset_band = max_band - band;
      for( unsigned int i = 0; i < 2*band+1; ++i) {

        unsigned int index = (i+offset_band)*(max_band+1) + band;

        sh_fortran_array[index].r = sh_coeffs(band,i);
        sh_fortran_array[index].i = 0.0;
      }
    }

  }

  //////////////////////////////////////////////////////////////////////////////
  //! Create SH coefficient vector from Fortran compatible representation
  //////////////////////////////////////////////////////////////////////////////
  template<class T>
  void
  convertShFromFortranArray( const unsigned int max_band,    
                             const std::vector< Complex<T> >& sh_fortran_array,
                             ShCoeffs<T>& sh_coeffs )
  {
    CUT_CONDITION( sh_coeffs.getMaxBand() == max_band);

    // swap content (respecting row / column interchange)
    for( unsigned int band = 0; band < max_band+1; ++band) {
      unsigned int offset_band = max_band - band;
      for( unsigned int i = 0; i < 2*band+1; ++i) {

        unsigned int index = (i+offset_band)*(max_band+1) + band;

        sh_coeffs(band,i) = sh_fortran_array[index].r;
      }
    }

  }


} // end namespace, unnamed

extern "C" {

  //! @param  alpha   rotation angle
  //! @param  nterms  order of the Spherical Harmonics expansion
  //! @param  m1      unused integer parameter  
  //! @param  m2      unused integer parameter  
  //! @param  mpole   input Spherical Harmonics coefficients
  //! @param  lmp     order of the expansion in mpole
  //! @param  marray2 rotated Spherical Harmonics coefficients
  //! @param  lmpn    order of the expansion in marray2 (upper bound)
  //! @param  w       work space (should be the same size as mpole and marray)
  //! @param  lw      size of workspace 
  //! @param  lused   work space used
  //! @note nterms, lmp, and lmpn should all be the same size
  void  rotviaproj_( const double* alpha, const unsigned int* nterms, 
                     const unsigned int* m1, const unsigned int* m2,
                     Complex<double>* mpole, const unsigned int* lmp,
                     Complex<double>* marray2, const unsigned int* lmpn,
                     double* w, int*lw, int* lused);

  // TODO: Port Greengard code to single precision.

} // end extern "C"

/*static*/ 
template<class T>
const std::string ShRotationMultipole<T>::class_name = "ShRotationMultipole";

////////////////////////////////////////////////////////////////////////////////
//! Constructor, default
////////////////////////////////////////////////////////////////////////////////
template<class T>
ShRotationMultipole<T>::ShRotationMultipole() :
  ShRotationBase<T>(),
  timer( 0)
{
  cutCreateTimer( &timer);
}

////////////////////////////////////////////////////////////////////////////////
//! Destructor
////////////////////////////////////////////////////////////////////////////////
template<class T>
ShRotationMultipole<T>::~ShRotationMultipole() { 

  cutDeleteTimer( timer);
}


////////////////////////////////////////////////////////////////////////////////
//! Initialize rotation, in particular load precomputed data
//! @param  max_band  maximum band to pre-load
//! @param  fname  file name containing the data needed for rotation
//!                the default name is 'data/pinchonNumMatJ.dat'
////////////////////////////////////////////////////////////////////////////////
template<class T>
void 
ShRotationMultipole<T>::preload( const unsigned int /*max_band*/, 
                                 std::string /*fname*/,
                                 const unsigned int /*os_factor*/)
{ }


////////////////////////////////////////////////////////////////////////////////
//! Rotate a signal represented in Spherical Harmonics
//! @return  Computation time of rotation matrix (ms).
//! @param  sh_coeffs  Spherical Harmonics coefficient vector (unrotated)
//! @param  sh_coeffs_rotated  Spherical Harmonics coefficient vector 
//!                            (after rotation)
//! @param  max_band   Maximum band of \a sh_coeffs to be rotated (that is 
//!                    \a sh_coeffs has to contain at least max_band bands)
//! @param  t1 / t2 / t3   ZYZ Euler rotation angles.
//! @param  num_iterations  Number of iterations (for computing rotation
//!                         matrices and measuring execution time)
////////////////////////////////////////////////////////////////////////////////
template<class T>
T
ShRotationMultipole<T>::rotate( const ShCoeffs<T>& sh_coeffs, 
                                ShCoeffs<T>& sh_coeffs_rotated, 
                                const unsigned int max_band,
                                const T& t1, const T& t2, const T& t3,
                                const unsigned int num_iterations,
                                const unsigned int /*os_factor = 100*/)
{
  // full rotation currently not implemented
  CUT_CONDITION( 0.0 == t1);
  CUT_CONDITION( 0.0 == t3);
  
  ShCoeffs<T> sh_coeffs_temp( sh_coeffs);
#if 1
  sh_coeffs_temp( 1,0) /= sqrt(2.0 / (4.0 * PI));
  sh_coeffs_temp( 1,1) /= sqrt(2.0 / (4.0 * PI));
  sh_coeffs_temp( 1,2) /= sqrt(2.0 / (4.0 * PI));
#endif

  // convert coefficients to Fortran format
  VecCd sh_coeffs_fortran;
  convertShToFortranArray( max_band, sh_coeffs_temp, sh_coeffs_fortran);
  VecCd sh_coeffs_rotated_fortran;
  resizeFortranArray( max_band, sh_coeffs_rotated_fortran);

  // temporary working memory
  int work_size = 20*max_band*max_band + 1000;
  std::vector<T> work( work_size, 0.0);
  int work_used = 0;

  // start timer
  cutStartTimer( timer);

  // perform rotation
  for( unsigned int i = 0; i < num_iterations; ++i) {
    rotviaproj_( &t2, &max_band, &max_band, &max_band,
                 &(sh_coeffs_fortran.front()), &max_band,
                 &(sh_coeffs_rotated_fortran.front()), &max_band,
                 &(work.front()), &work_size, &work_used);
  }
  CUT_CONDITION( work_used <= work_size);

  // stop time
  cutStopTimer( timer);

  //convert Fortran format back to our representation
  convertShFromFortranArray( max_band, sh_coeffs_rotated_fortran, 
                             sh_coeffs_rotated);

  // validate with Matlab
  // validateRotationMatlab( max_band, sh_coeffs, sh_coeffs_rotated, t1, t2, t3);

#if 1
  sh_coeffs_rotated( 1,0) *= sqrt(2.0 / (4.0 * PI));
  sh_coeffs_rotated( 1,1) *= sqrt(2.0 / (4.0 * PI));
  sh_coeffs_rotated( 1,2) *= sqrt(2.0 / (4.0 * PI));
#endif

#if 1
  std::cout << std::setprecision(7);
  std::cout << "Multipole :: sh_coeffs = " << sh_coeffs << '\n';
  std::cout << "Multipole :: sh_coeffs_rotated = " << sh_coeffs_rotated << '\n';
#endif

  // get timer value and then reset for next call
  T ret_val = cutGetTimerValue( timer) / ((T) num_iterations);
  cutResetTimer( timer);

  return ret_val;
}


////////////////////////////////////////////////////////////////////////////////
//! Rotate a signal represented in Spherical Harmonics
//! @return  Computation time of rotation matrix (ms).
//! @param  sh_coeffs  Spherical Harmonics coefficient vector (unrotated)
//! @param  sh_coeffs_rotated  Spherical Harmonics coefficient vector 
//!                            (after rotation)
//! @param  max_band   Maximum band of \a sh_coeffs to be rotated (that is 
//!                    \a sh_coeffs has to contain at least max_band bands)
//! @param  t1 / t2 / t3   ZYZ Euler rotation angles.
//! @param  num_iterations  Number of iterations (for computing rotation
//!                         matrices and measuring execution time)
////////////////////////////////////////////////////////////////////////////////
template<class T>
T
ShRotationMultipole<T>::rotatePreloaded( const ShCoeffs<T>& sh_coeffs, 
                                         ShCoeffs<T>& sh_coeffs_rotated, 
                                         const unsigned int max_band,
                                         const T& t1, const T& t2, const T& t3,
                                         const unsigned int num_iterations,
                                         const unsigned int os_factor)
{
  //std::cout << "WARNING :: ShRotationMultipole<T>::rotatePreloaded() : "
  //          << "No precomputation implemented." << std::endl; 

  return rotate( sh_coeffs, sh_coeffs_rotated, max_band,
                 t1, t2, t3, num_iterations, os_factor); 
}

////////////////////////////////////////////////////////////////////////////////
//! Get the name of the rotation algorithm
////////////////////////////////////////////////////////////////////////////////
template<class T>
const std::string& 
ShRotationMultipole<T>::name() const {
  return class_name;
}
  
