/// @file
////////////////////////////////////////////////////////////////////////////////
///
/// Copyright (C) 2009      University of Toronto
///
////////////////////////////////////////////////////////////////////////////////
///
///  module     : sh
///
///  author     : lessig@dgp.toronto.edu
///
///  project    : Spherical Harmonics rotation
///
///  description: Spherical Harmonics rotation following the paper by Pinchon
///               and Hoggan (Rotation matrices for real spherical harmonics:
///               general rotations of atomic orbitals in space-fixed axes).
///
///  license    : See README for details. Parts of the code have been obtained 
///               from Didier Pinchon.
///
////////////////////////////////////////////////////////////////////////////////

// includes, system
#include <fstream>
#include <iomanip>

// External functions in code provided by Pinchon and Hoggan
extern "C" {

  MatL* allocMatL( unsigned int max_band);

  //! Free matrix
  void freeMatL( MatL* mat);

  //! Read precomputed matrix data (for rotation around the Y axis)
  void  pinchonPreloadJ( const unsigned int max_band, MatL** matJ);

  //! Compute the rotation matrix for ZYZ rotation angles
  void   pinchonRotationMatrix( const unsigned int max_band, double alpha,
                                double beta, double gamma, const MatL* mat_j,
                                MatL* mat_r );


} // end extern "C"

// namespace, unnamed
namespace {



} // end namespace, unnamed

////////////////////////////////////////////////////////////////////////////////
/*static*/ 
template<class T>
const std::string ShRotationPinchon<T>::class_name = "ShRotationPinchon";

////////////////////////////////////////////////////////////////////////////////
//! Constructor, default
////////////////////////////////////////////////////////////////////////////////
template<class T>
ShRotationPinchon<T>::ShRotationPinchon() :
  ShRotationBase<T>(),
  mat_j_l(),
  coeffs_l(),
  res_l(),
  timer( 0)
{
  cutCreateTimer( &timer);
}

////////////////////////////////////////////////////////////////////////////////
//! Destructor
////////////////////////////////////////////////////////////////////////////////
template<class T>
ShRotationPinchon<T>::~ShRotationPinchon() { 

  // clean up J matrix
  freeMatL( mat_j_l);

  // cleanup rotation matrix for current band
  freeMatL( rot_l);

  // cleanup temporary vectors
  for( unsigned int band = 1; band < coeffs_l.size(); ++band) {
    gsl_vector_free( coeffs_l[band]);
    gsl_vector_free( res_l[band]);
  }

  cutDeleteTimer( timer);
}


////////////////////////////////////////////////////////////////////////////////
//! Initialize rotation, in particular load precomputed data
//! @param  max_band  maximum band to pre-load
//! @param  fname  file name containing the data needed for rotation
//!                the default name is 'data/pinchonNumMatJ.dat'
////////////////////////////////////////////////////////////////////////////////
template<class T>
void 
ShRotationPinchon<T>::preload( const unsigned int max_band) {

  // allocate MatL data
  mat_j_l = allocMatL( max_band);
  rot_l = allocMatL( max_band);

  // load precomputed data
  pinchonPreloadJ( max_band, &mat_j_l);

  // resize vectors 
  coeffs_l.resize( max_band+1);
  res_l.resize( max_band+1);

  // load J matrices
  for( unsigned int band = 1; band <= max_band; ++band) {
    
    unsigned int dim_band = 2*band+1;

    // allocate vectors so that this is not necessary at runtime
    coeffs_l[band] = gsl_vector_alloc( dim_band);
    res_l[band] = gsl_vector_alloc( dim_band);
  }
}

////////////////////////////////////////////////////////////////////////////////
//! Rotate a signal represented in Spherical Harmonics
//! @return  Computation time of rotation matrix (ms).
//! @param  sh_coeffs  Spherical Harmonics coefficient vector (unrotated)
//! @param  sh_coeffs_rotated  Spherical Harmonics coefficient vector 
//!                            (after rotation)
//! @param  max_band   Maximum band of \a sh_coeffs to be rotated (that is 
//!                    \a sh_coeffs has to contain at least max_band bands)
//! @param  t1 / t2 / t3   ZYZ Euler rotation angles.
//! @param  num_iterations  Number of iterations (for computing rotation
//!                         matrices and measuring execution time)
////////////////////////////////////////////////////////////////////////////////
template<class T>
T
ShRotationPinchon<T>::rotatePreloaded( const ShCoeffs<T>& sh_coeffs, 
                                       ShCoeffs<T>& sh_coeffs_rotated, 
                                       const unsigned int max_band,
                                       const T& t1, const T& t2, const T& t3,
                                       const unsigned int num_iterations)
{
  // start timing
  cutStartTimer( timer);
  
  // compute the rotation matrix
  // the Pinchon paper assumes an inverse order of rotation angles
  for( unsigned int i = 0; i < num_iterations; ++i) {
    
    // compute rotation matrix
    pinchonRotationMatrix( max_band, t3, t2, t1, mat_j_l, rot_l);

    // the DC component is invariant under rotation
    sh_coeffs_rotated(0,0) = sh_coeffs(0,0);

    // do for the remaining bands
    for( unsigned int band = 1; band <= max_band; ++band) {

      unsigned int dim_band = 2*band+1;

      // copy the basis function coefficients into the vector
      memcpy( gsl_vector_ptr( coeffs_l[band], 0), sh_coeffs.getAddress( band),
              sizeof(double) * dim_band);

      // compute the matrix vector product
      gsl_blas_dgemv( CblasNoTrans, 1.0, rot_l->lmat[band], 
                      coeffs_l[band], 0.0, res_l[band]);

      // copy the result back to the sh basis function coefficient vector
      memcpy( sh_coeffs_rotated.getAddress(band), 
              gsl_vector_ptr( res_l[band], 0), 
              sizeof(double) * dim_band);
    }

  }
  // stop timing
  cutStopTimer( timer);

  // get timer value and then reset for next call
  T ret_val = cutGetTimerValue( timer) / ((T) num_iterations);
  cutResetTimer( timer);

  return ret_val;
}

////////////////////////////////////////////////////////////////////////////////
//! Rotate a signal represented in Spherical Harmonics
//! @return  Computation time of rotation matrix (ms).
//! @param  sh_coeffs  Spherical Harmonics coefficient vector (unrotated)
//! @param  sh_coeffs_rotated  Spherical Harmonics coefficient vector 
//!                            (after rotation)
//! @param  max_band   Maximum band of \a sh_coeffs to be rotated (that is 
//!                    \a sh_coeffs has to contain at least max_band bands)
//! @param  t1 / t2 / t3   ZYZ Euler rotation angles.
//! @param  num_iterations  Number of iterations (for computing rotation
//!                         matrices and measuring execution time)
////////////////////////////////////////////////////////////////////////////////
template<class T>
T
ShRotationPinchon<T>::rotate( const ShCoeffs<T>& sh_coeffs, 
                              ShCoeffs<T>& sh_coeffs_rotated, 
                              const unsigned int max_band,
                              const T& t1, const T& t2, const T& t3,
                              const unsigned int num_iterations)
{
  // preload / precompute 
  preload( max_band);

  // perform rotation experiment
  return rotatePreloaded( sh_coeffs, sh_coeffs_rotated, max_band, 
                          t1, t2, t3, num_iterations);
}

////////////////////////////////////////////////////////////////////////////////
//! Get the name of the rotation algorithm
////////////////////////////////////////////////////////////////////////////////
template<class T>
const std::string& 
ShRotationPinchon<T>::name() const {
  return class_name;
}
  
