/// @file
////////////////////////////////////////////////////////////////////////////////
///
/// Copyright (C) 2009      University of Toronto
///
////////////////////////////////////////////////////////////////////////////////
///
///  module     : sh
///
///  author     : lessig@dgp.toronto.edu
///
///  project    : Spherical Harmonics rotation
///
///  description: Spherical Harmonics rotation following the paper by Pinchon
///               and Hoggan (Rotation matrices for real spherical harmonics:
///               general rotations of atomic orbitals in space-fixed axes).
///
///  license    : See README for details. Parts of the code have been obtained 
///               from Didier Pinchon.
///
////////////////////////////////////////////////////////////////////////////////

// includes, system
#include <fstream>
#include <iomanip>

// namespace, unnamed
namespace {

} // end namespace, unnamed

////////////////////////////////////////////////////////////////////////////////
/*static*/ 
template<class T>
const std::string ShRotationPinchon2<T>::class_name = "ShRotationPinchon2";

////////////////////////////////////////////////////////////////////////////////
//! Constructor, default
////////////////////////////////////////////////////////////////////////////////
template<class T>
ShRotationPinchon2<T>::ShRotationPinchon2() :
  ShRotationBase<T>(),
  mat_j_l(),
  coeffs_l(),
  res_l(),
  timer( 0)
{
  cutCreateTimer( &timer);
}

////////////////////////////////////////////////////////////////////////////////
//! Destructor
////////////////////////////////////////////////////////////////////////////////
template<class T>
ShRotationPinchon2<T>::~ShRotationPinchon2() { 

  // cleanup
  for( unsigned int band = 1; band < mat_j_l.size(); ++band) {
    
    // clean up J matrix
    gsl_matrix_free( mat_j_l[band]);

    // cleanup rotation matrix for current band
    gsl_matrix_free( rot_l[band]);

    // cleanup temporary vectors
    gsl_vector_free( coeffs_l[band]);
    gsl_vector_free( res_l[band]);
  }

  cutDeleteTimer( timer);
}


////////////////////////////////////////////////////////////////////////////////
//! Initialize rotation, in particular load precomputed data
//! @param  max_band  maximum band to pre-load
//! @param  fname  file name containing the data needed for rotation
//!                the default name is 'data/pinchonNumMatJ.dat'
////////////////////////////////////////////////////////////////////////////////
template<class T>
void 
ShRotationPinchon2<T>::preload( const unsigned int max_band) {

  mat_j_l.resize( max_band+1);
  rot_l.resize( max_band+1);
  coeffs_l.resize( max_band+1);
  res_l.resize( max_band+1);

  // load J matrices
  for( unsigned int band = 1; band <= max_band; ++band) {
    
    unsigned int dim_band = 2*band+1;

    // read J matrix for current band
    mat_j_l[band] = read_matJ( (int) band);

    // rotation matrix for current band
    rot_l[band] = gsl_matrix_calloc(dim_band,dim_band); 

    // allocate vectors so that this is not necessary at runtime
    coeffs_l[band] = gsl_vector_alloc( dim_band);
    res_l[band] = gsl_vector_alloc( dim_band);
  }
  // gsl_matrix_set( mat_j_l[1], 2, 2, 1.0);
}

////////////////////////////////////////////////////////////////////////////////
//! Rotate a signal represented in Spherical Harmonics
//! @return  Computation time of rotation matrix (ms).
//! @param  sh_coeffs  Spherical Harmonics coefficient vector (unrotated)
//! @param  sh_coeffs_rotated  Spherical Harmonics coefficient vector 
//!                            (after rotation)
//! @param  max_band   Maximum band of \a sh_coeffs to be rotated (that is 
//!                    \a sh_coeffs has to contain at least max_band bands)
//! @param  t1 / t2 / t3   ZYZ Euler rotation angles.
//! @param  num_iterations  Number of iterations (for computing rotation
//!                         matrices and measuring execution time)
////////////////////////////////////////////////////////////////////////////////
template<class T>
T
ShRotationPinchon2<T>::rotatePreloaded( const ShCoeffs<T>& sh_coeffs, 
                                       ShCoeffs<T>& sh_coeffs_rotated, 
                                       const unsigned int max_band,
                                       const T& t1, const T& t2, const T& t3,
                                       const unsigned int num_iterations)
{
  // start timing
  cutStartTimer( timer);
  
  // compute the rotation matrix
  // the Pinchon paper assumes an inverse order of rotation angles
  for( unsigned int i = 0; i < num_iterations; ++i) {
    
    // the DC component is invariant under rotation
    sh_coeffs_rotated(0,0) = sh_coeffs(0,0);

    // do for the remaining bands
    for( unsigned int band = 1; band <= max_band; ++band) {

      unsigned int dim_band = 2*band+1;

      // compute rotation matrix for the current band
      matrix_rotation( (int) band, t3, t2, t1,  
                       rot_l[band], mat_j_l[band]);

      // copy the basis function coefficients into the vector
      memcpy( gsl_vector_ptr( coeffs_l[band], 0), sh_coeffs.getAddress( band),
              sizeof(double) * dim_band);

      // compute the matrix vector product
      gsl_blas_dgemv( CblasNoTrans, 1.0, rot_l[band], 
                      coeffs_l[band], 0.0, res_l[band]);

      // copy the result back to the sh basis function coefficient vector
      memcpy( sh_coeffs_rotated.getAddress(band), 
              gsl_vector_ptr( res_l[band], 0), 
              sizeof(double) * dim_band);
    }

  }
  // stop timing
  cutStopTimer( timer);

  // get timer value and then reset for next call
  T ret_val = cutGetTimerValue( timer) / ((T) num_iterations);
  cutResetTimer( timer);

#if 0
  // std::cout << "Pinchon2 :: " << sh_coeffs << std::endl;
  std::cout << "Pinchon2 :: " << sh_coeffs_rotated << std::endl;
#endif

  return ret_val;
}

////////////////////////////////////////////////////////////////////////////////
//! Rotate a signal represented in Spherical Harmonics
//! @return  Computation time of rotation matrix (ms).
//! @param  sh_coeffs  Spherical Harmonics coefficient vector (unrotated)
//! @param  sh_coeffs_rotated  Spherical Harmonics coefficient vector 
//!                            (after rotation)
//! @param  max_band   Maximum band of \a sh_coeffs to be rotated (that is 
//!                    \a sh_coeffs has to contain at least max_band bands)
//! @param  t1 / t2 / t3   ZYZ Euler rotation angles.
//! @param  num_iterations  Number of iterations (for computing rotation
//!                         matrices and measuring execution time)
////////////////////////////////////////////////////////////////////////////////
template<class T>
T
ShRotationPinchon2<T>::rotate( const ShCoeffs<T>& sh_coeffs, 
                              ShCoeffs<T>& sh_coeffs_rotated, 
                              const unsigned int max_band,
                              const T& t1, const T& t2, const T& t3,
                              const unsigned int num_iterations)
{
  // preload / precompute 
  preload( max_band);

  // perform rotation experiment
  return rotatePreloaded( sh_coeffs, sh_coeffs_rotated, max_band, 
                          t1, t2, t3, num_iterations);
}

////////////////////////////////////////////////////////////////////////////////
//! Get the name of the rotation algorithm
////////////////////////////////////////////////////////////////////////////////
template<class T>
const std::string& 
ShRotationPinchon2<T>::name() const {
  return class_name;
}
  
