/// @file
////////////////////////////////////////////////////////////////////////////////
///
/// Copyright (C) 2009      University of Toronto
///
////////////////////////////////////////////////////////////////////////////////
///
///  module     : sh
///
///  author     : lessig@dgp.toronto.edu
///
///  project    : Spherical Harmonics rotation
///
///  description: Spherical Harmonics rotation following the paper by Pinchon
///               and Hoggan (Rotation matrices for real spherical harmonics:
///               general rotations of atomic orbitals in space-fixed axes).
///               Updated code provided by the authors July 2011.
///
///  license    : See README for details. Parts of the code have been obtained 
///               from Didier Pinchon.
///
////////////////////////////////////////////////////////////////////////////////

// includes, system
#include <fstream>
#include <iomanip>

// External functions in code provided by Pinchon and Hoggan
extern "C" {

  //! allocate matrix / vector
  MatL* allocMatL( unsigned int max_band);
  VecL* allocVecL( unsigned int max_band);

  //! free matrix / vector
  void freeMatL( MatL* mat);
  void freeVecL( VecL* mat);

  //! Read precomputed matrix data (for rotation around the Y axis)
  void  pinchonPreloadJ( const unsigned int max_band, MatL** matJ);

  //! Compute the rotation of the coefficient vector for ZYZ rotation angles
  void  pinchonRotateVector( const unsigned int max_band, double alpha, 
                             double beta, double gamma, VecL* sh_vec,
                             VecL* sh_vec_rot, const MatL* mat_j);

} // end extern "C"

////////////////////////////////////////////////////////////////////////////////
/*static*/ 
template<class T>
const std::string ShRotationPinchon3<T>::class_name = "ShRotationPinchon3";

////////////////////////////////////////////////////////////////////////////////
//! Constructor, default
////////////////////////////////////////////////////////////////////////////////
template<class T>
ShRotationPinchon3<T>::ShRotationPinchon3() :
  ShRotationBase<T>(),
  mat_j_l(),
  sh_vec(),
  sh_vec_rot(),
  timer( 0)
{
  cutCreateTimer( &timer);
}

////////////////////////////////////////////////////////////////////////////////
//! Destructor
////////////////////////////////////////////////////////////////////////////////
template<class T>
ShRotationPinchon3<T>::~ShRotationPinchon3() { 

  // clean up J matrix
  freeMatL( mat_j_l);

  // free vectors of basis function coefficients
  freeVecL( sh_vec);
  freeVecL( sh_vec_rot);

  // cleanup timer
  cutDeleteTimer( timer);
}


////////////////////////////////////////////////////////////////////////////////
//! Initialize rotation, in particular load precomputed data
//! @param  max_band  maximum band to pre-load
//! @param  fname  file name containing the data needed for rotation
//!                the default name is 'data/pinchonNumMatJ.dat'
////////////////////////////////////////////////////////////////////////////////
template<class T>
void 
ShRotationPinchon3<T>::preload( const unsigned int max_band) {

  // allocate MatL data
  mat_j_l = allocMatL( max_band);

  // load precomputed data
  pinchonPreloadJ( max_band, &mat_j_l);

  // resize vectors 
  sh_vec = allocVecL( max_band);
  sh_vec_rot = allocVecL( max_band);
}

////////////////////////////////////////////////////////////////////////////////
//! Rotate a signal represented in Spherical Harmonics
//! @return  Computation time of rotation matrix (ms).
//! @param  sh_coeffs  Spherical Harmonics coefficient vector (unrotated)
//! @param  sh_coeffs_rotated  Spherical Harmonics coefficient vector 
//!                            (after rotation)
//! @param  max_band   Maximum band of \a sh_coeffs to be rotated (that is 
//!                    \a sh_coeffs has to contain at least max_band bands)
//! @param  t1 / t2 / t3   ZYZ Euler rotation angles.
//! @param  num_iterations  Number of iterations (for computing rotation
//!                         matrices and measuring execution time)
////////////////////////////////////////////////////////////////////////////////
template<class T>
T
ShRotationPinchon3<T>::rotatePreloaded( const ShCoeffs<T>& sh_coeffs, 
                                       ShCoeffs<T>& sh_coeffs_rotated, 
                                       const unsigned int max_band,
                                       const T& t1, const T& t2, const T& t3,
                                       const unsigned int num_iterations)
{
  // copy input vector to the representation used by in the libcode
  for( unsigned int band = 1; band <= max_band; ++band) {
    memcpy( gsl_vector_ptr( sh_vec->lvec[band], 0), sh_coeffs.getAddress( band),
            sizeof(double) * (2*band+1));
  }

  // correct max_band of mat_j_l to match the current value
  mat_j_l->lmax = max_band;
  sh_vec->lmax = max_band;
  sh_vec_rot->lmax = max_band;

  // start timing
  cutStartTimer( timer);
  
  // compute the rotation matrix
  // the Pinchon paper assumes an inverse order of rotation angles
  for( unsigned int i = 0; i < num_iterations; ++i) {
    // compute rotation
    pinchonRotateVector( max_band, t3, t2, t1, sh_vec, sh_vec_rot, mat_j_l);
  }

  // stop timing
  cutStopTimer( timer);

  // copy result vector to our representation 
  for( unsigned int band = 1; band <= max_band; ++band) {
    memcpy( sh_coeffs_rotated.getAddress( band), 
            gsl_vector_ptr( sh_vec_rot->lvec[band], 0),
            sizeof(double) * (2*band+1));
  }
  // the DC component is invariant under rotation
  sh_coeffs_rotated(0,0) = sh_coeffs(0,0);

  // get timer value and then reset for next call
  T ret_val = cutGetTimerValue( timer) / ((T) num_iterations);
  cutResetTimer( timer);

  return ret_val;
}

////////////////////////////////////////////////////////////////////////////////
//! Rotate a signal represented in Spherical Harmonics
//! @return  Computation time of rotation matrix (ms).
//! @param  sh_coeffs  Spherical Harmonics coefficient vector (unrotated)
//! @param  sh_coeffs_rotated  Spherical Harmonics coefficient vector 
//!                            (after rotation)
//! @param  max_band   Maximum band of \a sh_coeffs to be rotated (that is 
//!                    \a sh_coeffs has to contain at least max_band bands)
//! @param  t1 / t2 / t3   ZYZ Euler rotation angles.
//! @param  num_iterations  Number of iterations (for computing rotation
//!                         matrices and measuring execution time)
////////////////////////////////////////////////////////////////////////////////
template<class T>
T
ShRotationPinchon3<T>::rotate( const ShCoeffs<T>& sh_coeffs, 
                              ShCoeffs<T>& sh_coeffs_rotated, 
                              const unsigned int max_band,
                              const T& t1, const T& t2, const T& t3,
                              const unsigned int num_iterations)
{
  // preload / precompute 
  preload( max_band);

  // perform rotation experiment
  return rotatePreloaded( sh_coeffs, sh_coeffs_rotated, max_band, 
                          t1, t2, t3, num_iterations);
}

////////////////////////////////////////////////////////////////////////////////
//! Get the name of the rotation algorithm
////////////////////////////////////////////////////////////////////////////////
template<class T>
const std::string& 
ShRotationPinchon3<T>::name() const {
  return class_name;
}
  
