/// @file
////////////////////////////////////////////////////////////////////////////////
///
/// Copyright (C) 2009      University of Toronto
///
////////////////////////////////////////////////////////////////////////////////
///
///  module     : sh
///
///  author     : lessig@dgp.toronto.edu
///
///  project    : Spherical Harmonics rotation
///
///  description: 
///
///  license    : See README for details. 
///
////////////////////////////////////////////////////////////////////////////////

// includes, system
#include <vector>
#include <iostream>

// includes, boost
#include <fftw3.h>

// includes, project
#include <point_s2.h>
#include <rotation.h>
#include <spectral_util.h>

using namespace Spectral;

// namespace, unnamed
namespace {

} // end namespace, unnamed

/*static*/ 
template<class T>
const std::string 
ShRotationRKHSSpectral<T>::class_name = "ShRotationRKHSSpectral";

////////////////////////////////////////////////////////////////////////////////
//! Constructor, default
////////////////////////////////////////////////////////////////////////////////
template<class T>
ShRotationRKHSSpectral<T>::ShRotationRKHSSpectral() :
  ShRotationBase<T>(),
  f_vals(),
  grad_proj(),
  timer( 0)
{
  cutCreateTimer( &timer);
}

////////////////////////////////////////////////////////////////////////////////
//! Destructor
////////////////////////////////////////////////////////////////////////////////
template<class T>
ShRotationRKHSSpectral<T>::~ShRotationRKHSSpectral() { 

  cutDeleteTimer( timer);
}


////////////////////////////////////////////////////////////////////////////////
//! Initialize rotation, in particular load precomputed data
//! @param  max_band  maximum band to pre-load
//! @param  fname  file name containing the data needed for rotation
//!                the default name is 'data/pinchonNumMatJ.dat'
////////////////////////////////////////////////////////////////////////////////
template<class T>
void 
ShRotationRKHSSpectral<T>::preload( const unsigned int max_band) {

  f_vals.resize( max_band+1);
  grad_proj.resize( max_band+1);

  for( unsigned int band = 0; band <= max_band ; ++band) {
    f_vals[band] = std::vector<T>( 2*max_band+1);
    grad_proj[band] = std::vector<T>( 2*max_band+1);
  }

}

////////////////////////////////////////////////////////////////////////////////
//! Rotate a signal represented in Spherical Harmonics
//! @note The implementation is similar to those by Greengard and Gimbutas.
//! @return  Computation time of rotation matrix (ms).
//! @param  sh_coeffs  Spherical Harmonics coefficient vector (unrotated)
//! @param  sh_coeffs_rotated  Spherical Harmonics coefficient vector 
//!                            (after rotation)
//! @param  max_band   Maximum band of \a sh_coeffs to be rotated (that is 
//!                    \a sh_coeffs has to contain at least max_band bands)
//! @param  t1 / t2 / t3   ZYZ Euler rotation angles.
//! @param  num_iterations  Number of iterations (for computing rotation
//!                         matrices and measuring execution time)
////////////////////////////////////////////////////////////////////////////////
template<class T>
T
ShRotationRKHSSpectral<T>::rotate( const ShCoeffs<T>& sh_coeffs_in, 
                             ShCoeffs<T>& sh_coeffs_r, 
                             const unsigned int max_band,
                             const T& t1, const T& t2, const T& t3,
                             const unsigned int num_iterations)
{
  // allocate temporaries

  // points on the equator
  std::vector< PointS2<T> > pts;
  const unsigned int num_samples = 2*max_band + 1;

  // allocate memory for discrete Fourier transform
  fftw_complex* f_vals_dft = 
    (fftw_complex*) fftw_malloc(sizeof(fftw_complex) * num_samples);
  fftw_complex* grad_proj_dft = 
    (fftw_complex*) fftw_malloc(sizeof(fftw_complex) * num_samples);
  T* f_vals_in = (T*) fftw_malloc(sizeof(T) * num_samples);
  T* grad_proj_in = (T*) fftw_malloc(sizeof(T) * num_samples);

  // determine optimized FFTW plans
  fftw_plan p1 = fftw_plan_dft_r2c_1d( num_samples, f_vals_in, f_vals_dft,
                                       FFTW_MEASURE);
  fftw_plan p2 = fftw_plan_dft_r2c_1d( num_samples, grad_proj_in, grad_proj_dft,
                                       FFTW_MEASURE); 


  //! temporaries needed for Legendre polynomials and its derivatives which are 
  //! needed for recovering the Spherical Harmonics coefficients
  std::vector<T> p_nm( max_band+1);
  std::vector<T> dp_nm( max_band+1);
  //! work array for evaluating Legendre polynomials
  std::vector<T> work( max_band+1);

  // create modifiable coefficient vector
  ShCoeffs<T> sh_coeffs( sh_coeffs_in);

  // construct *inverse* rotation matrix and rotate points
  Matrix R(3,3);
  Rotation::getRotationMatrixEulerZYZ( -t3, -t2, -t1, R);
    
  // start timer
  cutStartTimer( timer);

  for( unsigned int iters = 0; iters < num_iterations; ++iters) {

  // generate points on the equator and rotate the points
  createPointsEquiLatitudeR( num_samples, PI/2.0, pts, R);

  // compute sample values
  computeSampleValues( sh_coeffs, max_band, pts, R);

  // the zero-th band is truly invariant under rotation
  sh_coeffs_r(0,0) = sh_coeffs(0,0);

  // Helper for computing Legendre polynomials
  T p_nm_help = 1.0;

  // recover all bands
  for( unsigned int band = 1; band <= max_band; ++band) {

    // compute Fourier transform for function values and projected gradient
    
    // copy input date to fft arrays
    memcpy( f_vals_in, &(f_vals[band].front()), sizeof(T)*num_samples);
    memcpy( grad_proj_in, &(grad_proj[band].front()), sizeof(T)*num_samples);

    // execute fft
    fftw_execute(p1); 
    fftw_execute(p2); 

    // for our algorithm we essentially precompute this

    // compute scaling factors
    p_nm_help = Legendre::legendre( band, 0.0, 1.0, p_nm, dp_nm, p_nm_help);
    T scale_p_nm = 0.5 * sqrt((2*band+1) / (4*PI)) * num_samples;
    T scale_dp_nm = -0.5 * sqrt((2*band+1) / (4*PI)) * num_samples;

    // Finally recover SH coefficients. The discrete Fourier transform yields
    // band+1 complex coefficients.

    for( unsigned int i = 0; i < (band+1); ++i) {

      // Recover Spherical Harmonics basis function coefficients from the 
      // result of the discrete Fourier transform

      // "Filter" out the coefficients of interest
      // note that the sine value has to be written before the cosine one
      // since for i=0 the final address is the same (but the cosine value is
      // the correct one)
      if( 0 == (((band % 2) + (i % 2)) % 2)) {
        sh_coeffs_r(band,band-i) = valSin( i, f_vals_dft, p_nm, scale_p_nm);
        sh_coeffs_r(band,band+i) = valCos( i, f_vals_dft, p_nm, scale_p_nm);
      }
      else {
        sh_coeffs_r(band,band-i) = valSin(i,grad_proj_dft,dp_nm, scale_dp_nm);
        sh_coeffs_r(band,band+i) = valCos(i,grad_proj_dft,dp_nm, scale_dp_nm);
      }
    }
  }

  } // end iterations for accurate timing

  // get timer value and then reset for next call
  cutStopTimer( timer);
  T time = cutGetTimerValue( timer) / ((T) num_iterations);
  cutResetTimer( timer);

  // cleanup memory
  fftw_destroy_plan( p1);
  fftw_destroy_plan( p2);
  fftw_free( grad_proj_dft);
  fftw_free( grad_proj_in);
  fftw_free( f_vals_dft);
  fftw_free( f_vals_in);

  // validate with Matlab
  // validateRotationMatlab(max_band, sh_coeffs, sh_coeffs_rotated, t1, t2, t3);

#if 0
  std::cout << std::setprecision(7);
  std::cout << "Spectral :: sh_coeffs = " << sh_coeffs << '\n';
  std::cout << "Spectral :: sh_coeffs_rotated = " << sh_coeffs_rotated << '\n';
#endif

  return time;
}

////////////////////////////////////////////////////////////////////////////////
//! Rotate a signal represented in Spherical Harmonics
//! @return  Computation time of rotation matrix (ms).
//! @param  sh_coeffs  Spherical Harmonics coefficient vector (unrotated)
//! @param  sh_coeffs_rotated  Spherical Harmonics coefficient vector 
//!                            (after rotation)
//! @param  max_band   Maximum band of \a sh_coeffs to be rotated (that is 
//!                    \a sh_coeffs has to contain at least max_band bands)
//! @param  t1 / t2 / t3   ZYZ Euler rotation angles.
//! @param  num_iterations  Number of iterations (for computing rotation
//!                         matrices and measuring execution time)
////////////////////////////////////////////////////////////////////////////////
template<class T>
T
ShRotationRKHSSpectral<T>::rotatePreloaded( const ShCoeffs<T>& sh_coeffs, 
                                         ShCoeffs<T>& sh_coeffs_rotated, 
                                         const unsigned int max_band,
                                         const T& t1, const T& t2, const T& t3,
                                         const unsigned int num_iterations)
{
  //std::cout << "WARNING :: ShRotationRKHSSpectral<T>::rotatePreloaded() : "
  //          << "No precomputation implemented." << std::endl; 

  return rotate( sh_coeffs, sh_coeffs_rotated, max_band,
                 t1, t2, t3, num_iterations); 
}

////////////////////////////////////////////////////////////////////////////////
//! Compute sample values for the function and its derivative
//! @param max_band  maximum band
//! @param pts       sample points (rotated)
//! @param beta      rotation angle (a 
////////////////////////////////////////////////////////////////////////////////
template<class T>
void
ShRotationRKHSSpectral<T>::
computeSampleValues( ShCoeffs<T>& sh_coeffs, const unsigned int max_band, 
                     std::vector< PointS2<T> >& pts, const Matrix& R )
{
  // temporaries
  T f_band = 0.0;
  ShGrad  grad_s2_band( 0.0, 0.0);
  Vector  grad_r3_band( 3);

  // rotated theta = (0,0,1) vector
  Vector theta_orig( 3);
  Vector theta( 3);
  theta_orig(0) = 0.0;  theta_orig(1) = 0.0;  theta_orig(2) = 1.0;
  mult( R, theta_orig, theta);

  // for all sampling locations (the points on the equator with the *inverse*
  // rotation) applied
  for( unsigned int i = 0; i < pts.size(); ++i) {

    // compute values of Spherical Harmonics basis functions and gradient
    // at current point
    sh_coeffs.computeYGradY( max_band, pts[i]);

    // pre-compute values which are needed for computing the gradient in R^3

    T cos_theta = cos( pts[i].theta());  T sin_theta = sin( pts[i].theta());
    T cos_phi   = cos( pts[i].phi());    T sin_phi   = sin( pts[i].phi());

    Vector dtheta(3);
    dtheta(0) = cos_theta * cos_phi;  
    dtheta(1) = cos_theta * sin_phi;
    dtheta(2) = -sin_theta;

    Vector dphi(3);
    dphi(0) = -sin_phi;  dphi(1) = cos_phi;  dphi(2) = 0.0;

    // the dimension of the invariant subspace increases with the band and
    // hence also the number of vertices which have to be considered;
    // conversely, for lower bands we do not have to consider all vertices
    for( unsigned int band = 1; band <= max_band; ++band) {
     
      // compute function value and gradient at current sampling point 
      // for the signal restricted to the current band
      sh_coeffs.reconstructYGradYBandAtP( band, f_band, grad_s2_band);
      f_vals[band][i] = f_band;

      // compute gradient in R^3
      grad_r3_band =   grad_s2_band.first * dtheta 
                     + grad_s2_band.second / sin_theta * dphi;

      // compute projected gradient which is the theta component in the 
      // unrotated coordinate system
      grad_proj[band][i] = dot( grad_r3_band, theta);
    }
  }
}

////////////////////////////////////////////////////////////////////////////////
//! Get the name of the rotation algorithm
////////////////////////////////////////////////////////////////////////////////
template<class T>
const std::string& 
ShRotationRKHSSpectral<T>::name() const {
  return class_name;
}
  
