/// @file
////////////////////////////////////////////////////////////////////////////////
///
/// Copyright (C) 2009      University of Toronto
///
////////////////////////////////////////////////////////////////////////////////
///
///  module     : sh
///
///  author     : lessig@dgp.toronto.edu
///
///  project    : Spherical Harmonics rotation
///
///  description: Spherical Harmonics rotation using RKHS
///
///  license    : See README for details.
///
////////////////////////////////////////////////////////////////////////////////

// includes, system
#include <iomanip>

// includes, libraries
#include <boost/lexical_cast.hpp>

// includes, project
#include <rotation.h>

// namespace, unnamed
namespace {

  //! Legendre polynomials
  //! This function is based on gsl::gsl_sf_legendre_Pl_array
  template<class T, class TArray>
  void
  legendre( const unsigned int lmax, const T x, TArray& result_array) {

    /* upward recurrence: l P_l = (2l-1) z P_{l-1} - (l-1) P_{l-2} */

    T p_ellm2 = 1.0;    /* P_0(x) */
    T p_ellm1 = x;    /* P_1(x) */
    T p_ell = p_ellm1;

    result_array[0] = 1.0;
    result_array[1] = x;

    for( unsigned int ell=2; ell <= lmax; ell++){
      p_ell = (x*(2*ell-1)*p_ellm1 - (ell-1)*p_ellm2) / ell;
      p_ellm2 = p_ellm1;
      p_ellm1 = p_ell;
      result_array[ell] = p_ell;
    }
  }



} // end namespace, unnamed

////////////////////////////////////////////////////////////////////////////////
//! Constructor, default
////////////////////////////////////////////////////////////////////////////////
template<class T>
ShRotationRKHSZonal<T>::ShRotationRKHSZonal( const unsigned int oversampling_factor,
                                   const std::string path) :
  ShRotationBase<T>(),
  os_factor( oversampling_factor),
  num_bands_preloaded( 0),
  set_sampling_nested( false),
  set_sampling(),
  cbm(),
  cbm_t(),
  fvals(),
  path_data( path),
  leg(),
  timer( 0)
{
  class_name = std::string( "ShRotationRKHSZonal (");
  class_name += boost::lexical_cast<std::string>(os_factor);
  class_name += std::string(")");

  cutCreateTimer( &timer);
}

////////////////////////////////////////////////////////////////////////////////
//! Destructor
////////////////////////////////////////////////////////////////////////////////
template<class T>
ShRotationRKHSZonal<T>::~ShRotationRKHSZonal() {
  cutDeleteTimer(timer);
}

////////////////////////////////////////////////////////////////////////////////
//! Transpose change of basis matrix
////////////////////////////////////////////////////////////////////////////////
template<class T>
void
ShRotationRKHSZonal<T>::
cbmTranspose( const unsigned int band, const Matrix& cbm_c, Matrix& cbm_c_t) {

  for( unsigned int i = 0; i < 2*band+1; ++i) {      
    for (unsigned int j = 0; j < 2*band+1; ++j) {
      cbm_c_t(i,j) = cbm_c(j,i);
    }
  }

}

////////////////////////////////////////////////////////////////////////////////
//! Load rotation matrices and associated sets of sampling
//! @param  fname  file name containing the data needed for rotation
//!                the default name is 'sh_rotation_rkhs_b*.dat'
////////////////////////////////////////////////////////////////////////////////
template<class T>
void
ShRotationRKHSZonal<T>::preload( const unsigned int max_band) {

  // set default file name if none is specified
  if( 0 == path_data.length()) {
    path_data = "data/";
  }
  std::string fname( path_data);
  fname += "sh_rotation_rkhs_b";

  // set band 0 (the elements are only place-holders, for rotation the basis
  // function coefficients for band 0 can simply be copied
  set_sampling.push_back( std::vector< PointS2<T> >());
  Matrix  cbm_c( 1, 1);
  cbm_c(0,0) = 1;
  cbm.push_back( cbm_c);
  cbm_t.push_back( cbm_c);

  // do for all bands greater than 0
  for( unsigned int band = 1; band <= max_band; ++band) {

    //calculate dimensions of matrix and sampling set vector
    unsigned int dim_band = 2*band+1;
    unsigned int num_samples = numSamplesOversampling( band, os_factor);

    Matrix  cbm_c(dim_band,num_samples);
    assert( dim_band == num_samples);
    std::vector< PointS2<T> >  set_sampling_c(num_samples);

    loadCbm( band, fname, cbm_c, set_sampling_c);
    set_sampling.push_back( set_sampling_c);
    cbm.push_back( cbm_c);

    // calculate transpose of C matrix
    Matrix cbm_c_t(2*band+1,2*band+1);
    cbmTranspose( band, cbm_c, cbm_c_t);
    cbm_t.push_back( cbm_c_t);
  
  } // end reading input data

  num_bands_preloaded = max_band;

  // determine if the set of sampling is nested
  checkSetOfSamplingNested();

  // allocate memory for the temporaries
  prealloc( os_factor);

  // Variables needed for Zonal Harmonics evaluation

  // Legendre polynomials
  unsigned int num_samples = set_sampling[max_band].size();
  for( unsigned int i = 0; i < num_samples; ++i) {
    leg.push_back( std::vector<Vec>(num_samples, Vec(max_band+1,0.0)));
  }

  // rotated sampling points
  rotated_set_sampling.resize( num_samples);

}

////////////////////////////////////////////////////////////////////////////////
//! Load rotation matrices and associated sets of sampling
//! @param  fname  file name containing the data needed for rotation
//!                the default name is 'sh_rotation_rkhs_b*.dat'
////////////////////////////////////////////////////////////////////////////////
template<class T>
void
ShRotationRKHSZonal<T>::loadCbm( const unsigned int band, std::string fname, 
                                 Matrix& cbm_c,
                                 std::vector< PointS2<T> >& set_sampling_c) 
{
  std::string fname_c( fname);
  fname_c += boost::lexical_cast<std::string>( band);
  fname_c += "_os" + boost::lexical_cast<std::string>( os_factor);
  fname_c += ".bin";

  std::ifstream fs( fname_c.c_str());
  // check if opening the file succeeded
  if( ! fs.good()) {
    std::cerr << "WARNING: Failed to open input file for band "
              << band << ": '" << fname_c << "'.\n";
    return;
  }

  // read the data

  unsigned int num_samples = numSamplesOversampling(band,os_factor);

  unsigned int dim_band = 2 * band + 1;

  T *memblock = new T[dim_band * num_samples];

  fs.read((char*)memblock,sizeof(T)* num_samples * 3);

  T *block = memblock;

  for( unsigned int i = 0; i < num_samples; ++i) {
    set_sampling_c[i].x() = *block++;
    set_sampling_c[i].y() = *block++;
    set_sampling_c[i].z() = *block++;
  }

  fs.read((char*)memblock,sizeof(T)* dim_band * num_samples);

  block = memblock;
  for( unsigned int i = 0; i < dim_band; ++i) {
    for( unsigned int j = 0; j < num_samples; ++j) {
      cbm_c(i,j) = *block++;
    }
  }
  delete[] memblock;

}

////////////////////////////////////////////////////////////////////////////////
//! Check if the point set in \a set_sampling is nested and set
//! \a set_sampling_nested appropriately
////////////////////////////////////////////////////////////////////////////////
template<class T>
void
ShRotationRKHSZonal<T>::checkSetOfSamplingNested() {

  // Determine if the set of sampling is nested
  bool is_nested = true;
  for( unsigned int band = 1; band < num_bands_preloaded; ++band) {
    for( unsigned int i = 0; i < (2*band+1); ++i) {
      if( set_sampling[num_bands_preloaded-1][i] != set_sampling[band][i]) {
        is_nested = false;
        break;
      }
    }
    if( false == is_nested) {
      break;
    }
  }
  set_sampling_nested = is_nested;

#if DEBUG
  if( true == set_sampling_nested) {
    std::cout << "Set of sampling is nested.\n";
  }
  else {
    std::cout << "Set of sampling is NOT nested.\n";
  }
#endif

}

////////////////////////////////////////////////////////////////////////////////
//! Determine number of samples used in a band for a given oversampling rate
////////////////////////////////////////////////////////////////////////////////
template<class T>
unsigned int
ShRotationRKHSZonal<T>::numSamplesOversampling( unsigned int band, 
                                           unsigned int os_factor) 
{
  double os_factor_t  = ((double) os_factor) / 100.0;
  return ceil((2*band+1) * os_factor_t);
}

////////////////////////////////////////////////////////////////////////////////
//! Allocate memory for \a fvals, the temporary storage for function values
//! during rotation
////////////////////////////////////////////////////////////////////////////////
template<class T>
void
ShRotationRKHSZonal<T>::prealloc( const unsigned int os_factor) {

  fvals.resize( num_bands_preloaded+1);
  for( unsigned int band = 0; band < fvals.size(); ++band) {
    fvals[band] = Vector( numSamplesOversampling( band, os_factor));
  }
}

////////////////////////////////////////////////////////////////////////////////
//! Rotate a signal represented in Spherical Harmonics
//! @return  Computation time for function values (ms).
//! @param  sh_coeffs  Spherical Harmonics coefficient vector (unrotated)
//! @param  sh_coeffs_rotated  Spherical Harmonics coefficient vector
//!                            (after rotation)
//! @param  max_band   Maximum band of \a sh_coeffs to be rotated (that is
//!                    \a sh_coeffs has to contain at least max_band bands)
//! @param  t1 / t2 / t3  Euler ZYZ rotation angles
//! @param  num_iterations  Number of iterations (for computing rotation
//!                         matrices and measuring execution time)
////////////////////////////////////////////////////////////////////////////////
template<class T>
T
ShRotationRKHSZonal<T>::
rotatePreloaded( const ShCoeffs<T>& sh_coeffs,
                 ShCoeffs<T>& sh_coeffs_rotated,
                 const unsigned int max_band,
                 const T& t1, const T& t2, const T& t3,
                 const unsigned int iters)
{
  DEBUG_ASSERT( ( max_band <= num_bands_preloaded), "ShRotationRKHSZonal",
                  "rotatePreloaded()", "Not enough bands preloaded.");

  // internal coefficient vector for computations
  ShCoeffs<T>  sh_coeffs_r( sh_coeffs);

  // convert Spherical Harmonics to Zonal Harmonics coefficients
  std::vector<Vector> zh_coeffs( 1, Vector(1));
  for( unsigned int band = 1; band <= max_band; ++band) {

    // create temporary vectors
    Vector sh_coeffs_vect( 2*band+1, sh_coeffs_r.getAddress(band));
    zh_coeffs.push_back( Vector(2*band+1));

    // apply transformation matrix from SH basis to Zonal Harmonics(ZH) basis
    mult( cbm_t[band], sh_coeffs_vect, zh_coeffs[band]);
  }

  // compute rotation matrix
  Matrix R(3,3);
  // the rotation has to be applied inversely to the sample points
  Rotation::getRotationMatrixEulerZYZ( -t3, -t2, -t1, R);

  // start timer
  cutStartTimer( timer);

  for( unsigned int i = 0; i < iters; ++i) {

    // reconstruct the function at the rotated set of sampling
    computeFValsNestedZonal( zh_coeffs, max_band, R);

    // apply the change of basis matrix back to Spherical Harmonics

    // the coefficient of band zero is invariant under rotation
    sh_coeffs_rotated(0,0) = sh_coeffs(0,0);

    // iterate over all bands and apply change of basis matrices
    for( unsigned int band = 1; band <= max_band; ++band) {

      // create temporary vector
      Vector res( 2*band+1, sh_coeffs_rotated.getAddress(band));

      // perform matrix vector multiplication
      mult( cbm[band], fvals[band], res);
    }
  }

  // stop timer
  cutStopTimer( timer);

  // get timer value and then reset for next call
  T ret_val = cutGetTimerValue( timer) / ((T) iters);
  cutResetTimer( timer);

  return ret_val;
}

////////////////////////////////////////////////////////////////////////////////
//! Evaluate per band signals at rotated sample points for a nested set 
//! of sampling.
//! @param  sh_coeffs_r  Spherical Harmonics coefficient vector (unrotated)
//! @param  sh_coeffs_rotated  Spherical Harmonics coefficient vector
//!                            (after rotation)
//! @param  max_band   Maximum band of \a sh_coeffs to be rotated (that is
//!                    \a sh_coeffs has to contain at least max_band bands)
//! @param  R          rotation matrix
//! @param  iters      Number of iterations (for computing rotation
//!                    matrices and measuring execution time)
////////////////////////////////////////////////////////////////////////////////
template<class T>
void
ShRotationRKHSZonal<T>::
computeFValsNestedZonal( const std::vector<Vector>& zh_coeffs,
                         const unsigned int max_band, const Matrix& R) 
{
  unsigned int num_samples = set_sampling[max_band].size();

  // calculate rotated point set
  for (unsigned i = 0; i < num_samples; ++i) {
    rotated_set_sampling[i] = set_sampling[max_band][i];
    rotated_set_sampling[i].rotate( R);
  }

  // precompute dot product of sample points and rotated sample points
  // this will serve as argument to Legendre polynomial evaluation
  for (unsigned i = 0; i < num_samples; ++i) {
    for (unsigned j = 0; j < num_samples; ++j) {

      PointS2<T> p_orig;
      PointS2<T> p_rot;

      p_orig = set_sampling[max_band][j];
      p_rot = rotated_set_sampling[i];

      // dot product
      T cos_theta =  p_orig.dot(p_rot);
      // gsl_sf_legendre_Pl_array( max_band, cos_theta, &leg[i][j].front());
      legendre( max_band, cos_theta, leg[i][j]);
    }
  }

  // reconstruct f_l at all the points
  for( unsigned int band = 1; band <= max_band; ++band) {

    // perform reconstruction using the scalar product between basis function
    // values and ZH basis function coefficients
    for( unsigned int i = 0; i < (2*band+1); ++i) {

      fvals[band][i] = 0.0;
      T normalization = ((T) (2*band+1)) / (4.0*PI);

      for( unsigned int j = 0; j < (2*band+1); ++j) {
        fvals[band][i] += zh_coeffs[band][j] * leg[i][j][band];
      }
      fvals[band][i] *= normalization;
    }

    //normalization
  }

#if 0
    for( unsigned int i = 0; i < num_samples; ++i) {      
      for (unsigned int j = 0; j < num_samples; ++j) {
        // (*Pl[band])(i,j) = ((2.0*bt-1.0) * costheta(i,j) * 
        //    (*Pl[band-1])(i,j) - (bt-1.0) * (*Pl[band-2])(i,j)) / (bt);
      }	
    }
#endif
}

////////////////////////////////////////////////////////////////////////////////
//! Rotate a signal represented in Spherical Harmonics
//! @return  Computation time for function values (ms).
//! @param  sh_coeffs  Spherical Harmonics coefficient vector (unrotated)
//! @param  sh_coeffs_rotated  Spherical Harmonics coefficient vector
//!                            (after rotation)
//! @param  max_band   Maximum band of \a sh_coeffs to be rotated (that is
//!                    \a sh_coeffs has to contain at least max_band bands)
//! @param  t1 / t2 / t3  Euler ZYZ rotation angles
//! @param  num_iterations  Number of iterations (for computing rotation
//!                         matrices and measuring execution time)
////////////////////////////////////////////////////////////////////////////////
template<class T>
T
ShRotationRKHSZonal<T>::rotate( const ShCoeffs<T>& sh_coeffs,
                           ShCoeffs<T>& sh_coeffs_rotated,
                           const unsigned int max_band,
                           const T& t1, const T& t2, const T& t3,
                           const unsigned int num_iterations)
{
  preload( max_band);

  return rotatePreloaded( sh_coeffs, sh_coeffs_rotated, max_band, t1, t2, t3, 
                          num_iterations);
}

////////////////////////////////////////////////////////////////////////////////
//! Get the name of the rotation algorithm
////////////////////////////////////////////////////////////////////////////////
template<class T>
const std::string&
ShRotationRKHSZonal<T>::name() const {
  return class_name;
}
  
