# mkMatJFast.mpl
# Didier Pinchon, December 2009

# Computing J_l from J_{l-1} by recurrence and
# storing J_l by blocks in exactMatJ-l.dat

with(LinearAlgebra):

# Quadratic numbers of the form a*sqrt(b)/2^c ,c >=0, and (a,2)=1 when c>0,
# are represented by the record [a,c,b] (pure binary quadratic numbers)
expo2 := proc(n)
  local lf;
  lf := ifactors(n);
  if not(lf[1]=1) then RETURN("1:not a power of 2"); fi;
  if nops(lf[2])=0 then RETURN(0); fi;
  if not(nops(lf[2])=1 and lf[2,1,1]=2) then RETURN("2:not a power of 2"); fi;
  lf[2,1,2];
end:
  
pureQuadratic2rec := proc(x)
  local lf,f,den,a,b, check,pow2;
  if x=0 then RETURN([0,1,0]); fi;
  if evalf(x)<0 then a:=-1; else a:=+1; fi;
  den := denom(x);
  pow2 := expo2(den);
  lf := ifactors(numer(x)^2);
  b := 1;
  for f in lf[2] do
    a := a*f[1]^iquo(f[2],2);
    b := b*f[1]^irem(f[2],2);
  od;
  check := simplify(x - a*sqrt(b)/2^pow2);
  if not(check=0) then ERROR("error in decompPureQuadratic!"); fi;
  [a,pow2,b];
end:

rec2pureQuadratic := proc(rec) rec[1]*sqrt(rec[3])/2^rec[2]; end proc:

readPureQuadratic := proc(f)
  local st,rec;
  st := readline(f);
  rec := sscanf(st,"%d %d %d");
  rec2pureQuadratic(rec);
end proc:

printPureQuadratic := proc(f,x)
  fprintf(f,"%d %d %d\n",op(pureQuadratic2rec(x)));
end proc:
       
# Permutation Q_l
sigma := proc(l,i)
  if i<=l then 2*i; else 2*(i-l)-1; end if;
end proc:

# Input from Jl files
readJfile := proc(direc,l)
  local fname,st,f,J,k,r,i,j;
  fname := cat(direc,"/exactMatJ-",l,".dat"); # For Unix systems
  f := fopen(fname,READ);
  st := readline(f);
  k := iquo(l,2,'r');
  J := Matrix(2*l+1,2*l+1,0);
  for i to k do
    for j to i-1 do
       J[sigma(l,i),sigma(l,j)] := readPureQuadratic(f);
       J[sigma(l,j),sigma(l,i)] := J[sigma(l,i),sigma(l,j)];
    end do;
    J[sigma(l,i),sigma(l,i)] := readPureQuadratic(f);
  end do;
  for i to k+r do
    for j to k+r do
      J[sigma(l,l+i),sigma(l,k+j)] := readPureQuadratic(f);
      J[sigma(l,k+j),sigma(l,l+i)] := J[sigma(l,l+i),sigma(l,k+j)];
    end do;
  end do;
  for i to k+1 do
    for j to i-1 do
       J[sigma(l,l+k+r+i),sigma(l,l+k+r+j)] := readPureQuadratic(f);
       J[sigma(l,l+k+r+j),sigma(l,l+k+r+i)] := J[sigma(l,l+k+r+i),sigma(l,l+k+r+j)];
    end do;
    J[sigma(l,l+k+r+i),sigma(l,l+k+r+i)] := readPureQuadratic(f);
  end do;
  fclose(f);
  J;
end proc:

# Output to Jl files
printJfile := proc(direc,J)
  local fname,l,f,k,r,i,j;
  l := iquo(RowDimension(J),2);
  fname := cat(direc,"/exactMatJ-",l,".dat"); # For Unix systems
  f := fopen(fname,WRITE);
  fprintf(f,"###l= %d\n",l);
  k := iquo(l,2,'r');
  for i to k do
    for j to i do
       printPureQuadratic(f, J[sigma(l,i),sigma(l,j)]);
    end do;
  end do;
  for i to k+r do
    for j to k+r do
      printPureQuadratic(f, J[sigma(l,l+i),sigma(l,k+j)]);
    end do;
  end do;
  for i to k+1 do
    for j to i do
       printPureQuadratic(f, J[sigma(l,l+k+r+i),sigma(l,l+k+r+j)]);
    end do;
  end do;
  fclose(f);
end proc:

# Recurrence for J_l
eps1 := proc(l,k)
  if k<=l    then return(-1);
  elif k=l+1 then return(-sqrt(2));
  elif k=l+2 then return(sqrt(2));
  end if;
  1;
end proc:

eps2 := proc(l,k)
  if k<=l+1 then return(-1);
  elif k=l+2 or k=l+3 then return(0);
  end if;
  1;
end proc:

nextJl := proc(Jl)
  local l,J,k,j;
  l := iquo(RowDimension(Jl),2);
  J := Matrix(2*l+3,2*l+3,0);
  for k to 2*l+1 do
    for j to 2*l+1 do
       J[k,j+1] := eps1(l,k)*sqrt((2*l+2-k)*(2*l+3-k))*Jl[2*l+2-k,j];
    end do;
  end do;
  for k from 3 to 2*l+3 do
    for j to 2*l+1 do
      J[k,j+1] := J[k,j+1] + eps2(l,k)*sqrt((k-2)*(k-1))*Jl[2*l+4-k,j];
    end do;
  end do;
  for k to 2*l+3 do
    for j to 2*l+1 do
       J[k,j+1] := combine(simplify(J[k,j+1]/2/sqrt(j*(2*l+2-j)) ));
    end do;
  end do;
  for k from 2 to 2*l+2 do
    J[k,1] := J[1,k];
    J[k,2*l+3] := J[2*l+3,k];
  end do;
  J[2*l+3,2*l+3] := 1/2^l;
  J;
end proc:
